"""
Basic Amplitude Calculations.
A partial wave analysis process has following structure:

DecayGroup: addition (+)
    DecayChain: multiplication (x)
        Decay, Particle(Propagator)

"""

import contextlib
import functools
import inspect
import warnings
from itertools import combinations
from pprint import pprint

import numpy as np
import sympy as sym

from tf_pwa.breit_wigner import BW, BWR, Bprime, Bprime_q2
from tf_pwa.cg import cg_coef
from tf_pwa.config import get_config, regist_config, temp_config
from tf_pwa.data import data_map, data_shape, split_generator
from tf_pwa.dec_parser import load_dec_file
from tf_pwa.dfun import get_D_matrix_lambda
from tf_pwa.einsum import einsum
from tf_pwa.particle import DEFAULT_DECAY, BaseParticle, Decay
from tf_pwa.particle import DecayChain as BaseDecayChain
from tf_pwa.particle import DecayGroup as BaseDecayGroup
from tf_pwa.particle import _spin_int, _spin_range, split_particle_type
from tf_pwa.tensorflow_wrapper import tf
from tf_pwa.variable import Variable, VarsManager

# from pysnooper import snoop


PARTICLE_MODEL = "particle_model"
regist_config(PARTICLE_MODEL, {})
DECAY_MODEL = "decay_model"
regist_config(DECAY_MODEL, {})
DECAY_CHAIN_MODEL = "decay_chain_model"
regist_config(DECAY_CHAIN_MODEL, {})


def register_particle(name=None, f=None):
    """register a particle model

    :params name: model name used in configuration
    :params f: Model class
    """

    def regist(g):
        if name is None:
            my_name = g.__name__
        else:
            my_name = name
        config = get_config(PARTICLE_MODEL)
        if my_name in config:
            warnings.warn("Override model {}".format(my_name))
        config[my_name] = g
        g.model_name = my_name
        return g

    if f is None:
        return regist
    return regist(f)


def register_decay(name=None, num_outs=2, f=None):
    """register a decay model

    :params name: model name used in configuration
    :params f: Model class
    """

    def regist(g):
        if name is None:
            my_name = g.__name__
        else:
            my_name = name
        config = get_config(DECAY_MODEL)
        id_ = (num_outs, my_name)
        if id_ in config:
            warnings.warn("Override deccay model {}".format(my_name))
        config[id_] = g
        g.model_name = my_name
        return g

    if f is None:
        return regist
    return regist(f)


def register_decay_chain(name=None, f=None):
    """register a decay model

    :params name: model name used in configuration
    :params f: Model class
    """

    def regist(g):
        if name is None:
            my_name = g.__name__
        else:
            my_name = name
        config = get_config(DECAY_CHAIN_MODEL)
        id_ = my_name
        if id_ in config:
            warnings.warn("Override deccay model {}".format(my_name))
        config[id_] = g
        g.model_name = my_name
        return g

    if f is None:
        return regist
    return regist(f)


regist_particle = register_particle
regist_decay = register_decay


def get_particle_model(name):
    all_model = get_config(PARTICLE_MODEL)
    return all_model.get(name, None)


def get_particle_model_name(p):
    all_model = get_config(PARTICLE_MODEL)
    for k, v in all_model.items():
        if type(p) is v:
            return k
    return str(type(p))


def get_particle(*args, model="default", **kwargs):
    """method for getting particle of model"""
    if isinstance(model, dict):
        model_class = trans_model(model)
    else:
        model_class = get_particle_model(model)
    if model_class is None:
        warnings.warn(
            "No model named {} found, use default instead.".format(model)
        )
        model_class = get_particle_model("default")
    return model_class(*args, **kwargs)


def trans_model(model):
    expr = model.get("expr")
    expr = sym.simplify(expr)
    var = {str(k): str(k) for k in expr.free_symbols}
    var.update(model.get("where", {}))
    model_name = []
    for k, v in var.items():
        if isinstance(v, str):
            model_name.append((k, v))
    assert len(model_name) == 1
    expr = sym.simplify(expr)
    var_name, name = model_name.pop()
    expr2 = expr.subs({k: v for k, v in var.items() if k != var_name})
    assert len(expr2.free_symbols) == 1, str(expr2)
    fun = sym.lambdify((var_name,), expr2, "tensorflow")
    base_model = get_particle_model(name)

    class _TempModel(base_model):
        _from_trans = True

        def get_amp(self, *args, **kwargs):
            amp = super().get_amp(*args, **kwargs)
            return fun(amp)

    return _TempModel


def get_decay(core, outs, **kwargs):
    """method for getting decay of model"""
    num_outs = len(outs)

    prod_params = {}
    for i in outs:
        prod_params.update(getattr(i, "production_params", {}))

    decay_params = getattr(core, "decay_params", {})

    new_kwargs = {**prod_params, **decay_params, **kwargs}

    model = new_kwargs.get("model", "default")
    id_ = (num_outs, model)

    return get_config(DECAY_MODEL)[id_](core, outs, **new_kwargs)


def get_decay_chain(decays, **kwargs):
    """method for getting decay of model"""
    decay_params = {}
    for i in decays:
        decay_params.update(getattr(i, "decay_chain_params", {}))

    new_kwargs = {**decay_params, **kwargs}

    model = new_kwargs.pop("model", "default")

    return get_config(DECAY_CHAIN_MODEL)[model](decays, **new_kwargs)


def data_device(data):
    def get_device(dat):
        if hasattr(dat, "device"):
            return dat.device
        return None

    pprint(data_map(data, get_device))
    return data


def get_name(self, names):
    name = (
        (str(self) + "_" + names)
        .replace(":", "/")
        .replace("+", ".")
        .replace(",", "")
        .replace("[", "")
        .replace("]", "")
        .replace(" ", "")
    )
    return name


def _add_var(self, names, is_complex=False, shape=(), **kwargs):
    name = get_name(self, names)
    return Variable(name, shape, is_complex, **kwargs)


class AmpBase(object):
    """Base class for amplitude """

    def add_var(self, names, is_complex=False, shape=(), **kwargs):
        """
        default add_var method
        """
        if not hasattr(self, "_variables_map"):
            self._variables_map = {}
        name = self.get_variable_name(names)
        var = Variable(name, shape, is_complex, **kwargs)
        self._variables_map[names] = var
        return var

    def get_var(self, name):
        return getattr(self, "_variables_map", {}).get(name)

    def get_variable_name(self, name=""):
        return get_name(self, name)

    def amp_shape(self):
        raise NotImplementedError

    def get_factor_variable(self):
        return []


@contextlib.contextmanager
def variable_scope(vm=None):
    """variabel name scope"""
    if vm is None:
        vm = VarsManager(dtype=get_config("dtype"))
    with temp_config("vm", vm):
        yield vm


def simple_deepcopy(dic):
    if isinstance(dic, dict):
        return {k: simple_deepcopy(v) for k, v in dic.items()}
    if isinstance(dic, list):
        return [simple_deepcopy(v) for v in dic]
    if isinstance(dic, tuple):
        return tuple([simple_deepcopy(v) for v in dic])
    return dic


def simple_cache_fun(f):
    name = "simple_cached_" + f.__name__

    @functools.wraps(f)
    def g(self, *args, **kwargs):
        if not hasattr(self, name):
            setattr(self, name, f(self, *args, **kwargs))
        return getattr(self, name)

    return g


def get_relative_p(m_0, m_1, m_2):
    """relative momentum for 0 -> 1 + 2"""
    M12S = m_1 + m_2
    M12D = m_1 - m_2
    if hasattr(M12S, "dtype"):
        m_0 = tf.convert_to_tensor(m_0, dtype=M12S.dtype)
    m_eff = tf.where(m_0 > M12S, m_0, M12S)
    p = (m_eff - M12S) * (m_eff + M12S) * (m_eff - M12D) * (m_eff + M12D)
    # if p is negative, which results from bad data, the return value is 0.0
    # print("p", tf.where(p==0), m_0, m_1, m_2)
    return tf.sqrt(p) / (2 * m_eff)


def get_relative_p2(m_0, m_1, m_2):
    """relative momentum for 0 -> 1 + 2"""
    M12S = m_1 + m_2
    M12D = m_1 - m_2
    if hasattr(M12S, "dtype"):
        m_0 = tf.convert_to_tensor(m_0, dtype=M12S.dtype)
    # m_eff = tf.where(m_0 > M12S, m_0, M12S)
    p = (m_0 - M12S) * (m_0 + M12S) * (m_0 - M12D) * (m_0 + M12D)
    # if p is negative, which results from bad data, the return value is 0.0
    # print("p", tf.where(p==0), m_0, m_1, m_2)
    return p / (2 * m_0) ** 2


def _ad_hoc(m0, m_max, m_min):
    r"""ad-hoc formula

    .. math::
        m_0^{eff} = m^{min} + \frac{m^{max} - m^{min}}{2}(1+tanh \frac{m_0 - \frac{m^{max} + m^{min}}{2}}{m^{max} - m^{min}})

    """
    k = (m_max - m_min) / 2
    m_eff = k * (1 + tf.tanh((2 * m0 - (m_max + m_min)) / k))
    return m_eff + m_min


@regist_particle("BWR")
@regist_particle("default")
class Particle(BaseParticle, AmpBase):
    """
    .. math::
        R(m) = \\frac{1}{m_0^2 - m^2 - i m_0 \\Gamma(m)}

    """

    def __init__(self, *args, running_width=True, bw_l=None, **kwargs):
        super(Particle, self).__init__(*args, **kwargs)
        self.running_width = running_width
        self.bw_l = bw_l

    def init_params(self):
        self.d = 3.0
        if self.mass is None:
            self.mass = self.add_var("mass", fix=True)
            # print("$$$$$",self.mass)
        else:
            if not isinstance(self.mass, Variable):
                self.mass = self.add_var("mass", value=self.mass, fix=True)
        if self.width is not None:
            if not isinstance(self.width, Variable):
                self.width = self.add_var("width", value=self.width, fix=True)

    def get_amp(self, data, data_c, **kwargs):
        mass = self.get_mass()
        width = self.get_width()
        if width is None:
            return tf.ones_like(data["m"])
        if not self.running_width:
            ret = BW(data["m"], mass, width)
        else:
            q = data_c["|q|"]
            q0 = data_c["|q0|"]
            if self.bw_l is None:
                decay = self.decay[0]
                self.bw_l = min(decay.get_l_list())
            ret = BWR(data["m"], mass, width, q, q0, self.bw_l, self.d)
            # ret = tf.where(q0 > 0, ret, tf.zeros_like(ret))
            # ret = tf.where(q > 0, ret, tf.zeros_like(ret))
        return ret

    def amp_shape(self):
        return ()

    def get_mass(self):
        if self.mass is None:
            warnings.warn(
                f"The mass of {self} is None, may be you should calculate amplitude first to infer mass"
            )
        if callable(self.mass):
            return self.mass()
        return self.mass

    def get_width(self):
        if callable(self.width):
            return self.width()
        return self.width


@regist_particle("x")
class ParticleX(BaseParticle, AmpBase):
    """simple particle model for mass, (used in expr)

    .. math::
        R(m) = m

    """

    def __call__(self, m):
        return self.get_amp({"m": m})

    def get_amp(self, data, *args, **kwargs):
        m = data["m"]
        zeros = tf.zeros_like(m)
        return tf.complex(m, zeros)


class SimpleResonances(Particle):
    def __init__(self, *args, **kwargs):
        self.params = {}
        super(SimpleResonances, self).__init__(*args, **kwargs)

    def __call__(self, m, m0=None, g0=None, q=None, q0=None, **kwargs):
        raise NotImplementedError

    def get_amp(self, *args, **kwargs):
        m = args[0]["m"]
        q, q0 = None, None
        if len(args) >= 2:
            q = args[1].get("|q|", 1.0)
            q0 = args[1].get("|q0|", 1.0)
        m0 = self.get_mass()
        g0 = self.get_width()
        return self(m, m0=m0, g0=g0, q=q, q0=q0, **kwargs)


class FloatParams(float):
    pass


def simple_resonance(name, fun=None, params=None):
    """convert simple fun f(m) into a resonances model

    :params name: model name used in configuration
    :params fun: Model function
    :params params: arguments name list for parameters

    """

    if params is None:
        params = {}

    def _wrapper(f):
        argspec = inspect.getfullargspec(f)
        args = argspec.args
        if argspec.defaults is None:
            defaults = {}
        else:
            defaults = dict(zip(argspec.args[::-1], argspec.defaults[::-1]))

        @register_particle(name)
        class _R(SimpleResonances):
            def init_params(self):
                if "m0" in argspec.args and "g0" in argspec.args:
                    super(_R, self).init_params()
                self.params = {}
                for i in argspec.args:
                    tp = argspec.annotations.get(i, None)
                    if i in params or tp is FloatParams:
                        val = getattr(self, i, defaults.get(i, None))
                        if val is None:
                            self.params[i] = self.add_var(i)
                        else:
                            self.params[i] = self.add_var(
                                i, value=val, fix=True
                            )

            def __call__(self, m, **kwargs):
                my_kwargs = {}
                for i in argspec.args:
                    if i in kwargs:
                        my_kwargs[i] = kwargs[i]
                    elif i in self.params:
                        my_kwargs[i] = self.params[i]()
                    elif hasattr(self, i):
                        my_kwargs[i] = getattr(self, i)
                ret = f(m, **my_kwargs)
                return tf.cast(ret, tf.complex128)

            __call__.__doc__ = f.__doc__

        _R.get_amp.__doc__ = f.__doc__
        return _R

    if fun is None:
        return _wrapper
    return _wrapper(fun)


class AmpDecay(Decay, AmpBase):
    """base class for decay with amplitude"""

    def amp_shape(self):
        ret = [len(self.core.spins)]
        for i in self.outs:
            ret.append(len(i.spins))
        return tuple(ret)

    # @simple_cache_fun
    def amp_index(self, base_map):
        ret = [base_map[self.core]]
        for i in self.outs:
            ret.append(base_map[i])
        return ret


@regist_decay("default")
@regist_decay("gls-bf")
class HelicityDecay(AmpDecay):
    """default decay model"""

    def __init__(
        self,
        *args,
        has_barrier_factor=True,
        l_list=None,
        barrier_factor_mass=False,
        has_bprime=True,
        aligned=False,
        allow_cc=True,
        ls_list=None,
        barrier_factor_norm=False,
        params_polar=None,
        **kwargs
    ):
        super(HelicityDecay, self).__init__(*args, **kwargs)
        self.has_barrier_factor = has_barrier_factor
        self.l_list = l_list
        self.barrier_factor_mass = barrier_factor_mass
        self.has_bprime = has_bprime
        self.aligned = aligned
        self.allow_cc = allow_cc
        self.single_gls = False
        self.ls_index = None
        self.total_ls = None
        self.barrier_factor_norm = barrier_factor_norm
        self.ls_list = None
        if ls_list is not None:
            self.ls_list = tuple([tuple(i) for i in ls_list])
        self.params_polar = params_polar

    def check_valid_jp(self):
        if len(self.get_ls_list()) == 0:
            if not self.p_break:
                raise ValueError(
                    """invalid spin parity for {}, maybe you should set `p_break: True` for weak decay""".format(
                        self
                    )
                )
            raise ValueError("invalid spin parity for {}".format(self))

    def set_ls(self, ls):
        if self.total_ls is None:
            self.total_ls = self.get_ls_list()
        self.ls_list = tuple([tuple(i) for i in ls])
        self.single_gls = len(ls) == 1
        # print(self, "total_ls: ", self.total_ls)
        total_ls = self.total_ls
        if len(total_ls) == len(ls):
            self.ls_index = None
            return
        self.ls_index = []
        for i in self.ls_list:
            self.ls_index.append(total_ls.index(i))

    def init_params(self):
        self.d = 3.0
        ls = self.get_ls_list()
        self.g_ls = self.add_var(
            "g_ls", is_complex=True, polar=self.params_polar, shape=(len(ls),)
        )
        try:
            self.g_ls.set_fix_idx(fix_idx=0, fix_vals=(1.0, 0.0))
        except Exception as e:
            print(e, self, self.get_ls_list())

    def get_factor_variable(self):
        return [(self.g_ls,)]

    def _get_particle_mass(self, p, data, from_data=False):
        if from_data:
            return data[p]["m"]
        if p.mass is None:
            p.mass = tf.reduce_mean(data[p]["m"])
        return p.get_mass()

    def get_relative_momentum(self, data, from_data=False):
        """"""

        _get_mass = lambda p: self._get_particle_mass(p, data, from_data)

        m0 = _get_mass(self.core)
        m1 = _get_mass(self.outs[0])
        m2 = _get_mass(self.outs[1])
        return get_relative_p(m0, m1, m2)

    def get_relative_momentum2(self, data, from_data=False):
        """"""

        _get_mass = lambda p: self._get_particle_mass(p, data, from_data)

        m0 = _get_mass(self.core)
        m1 = _get_mass(self.outs[0])
        m2 = _get_mass(self.outs[1])
        return get_relative_p2(m0, m1, m2)

    def get_cg_matrix(self):
        ls = self.get_ls_list()
        return self._get_cg_matrix(ls)

    @functools.lru_cache()
    def _get_cg_matrix(self, ls):  # CG factor inside H
        """
        [(l,s),(lambda_b,lambda_c)]

        .. math::
          \\sqrt{\\frac{ 2 l + 1 }{ 2 j_a + 1 }}
          \\langle j_b, j_c, \\lambda_b, - \\lambda_c | s, \\lambda_b - \\lambda_c \\rangle
          \\langle l, s, 0, \\lambda_b - \\lambda_c | j_a, \\lambda_b - \\lambda_c \\rangle
        """
        m = len(ls)
        ja = self.core.J
        jb = self.outs[0].J
        jc = self.outs[1].J
        n = _spin_int(2 * jb + 1), _spin_int(2 * jc + 1)
        ret = np.zeros(shape=(m, *n))
        for i, ls_i in enumerate(ls):
            l, s = ls_i
            for i1, lambda_b in enumerate(_spin_range(-jb, jb)):
                for i2, lambda_c in enumerate(_spin_range(-jc, jc)):
                    ret[i][i1][i2] = (
                        np.sqrt((2 * l + 1) / (2 * ja + 1))
                        * cg_coef(
                            jb, jc, lambda_b, -lambda_c, s, lambda_b - lambda_c
                        )
                        * cg_coef(
                            l,
                            s,
                            0,
                            lambda_b - lambda_c,
                            ja,
                            lambda_b - lambda_c,
                        )
                    )
        return tf.convert_to_tensor(ret)

    def get_helicity_amp(self, data, data_p, **kwargs):
        m_dep = self.get_ls_amp(data, data_p, **kwargs)
        cg_trans = tf.cast(self.get_cg_matrix(), m_dep.dtype)
        n_ls = len(self.get_ls_list())
        m_dep = tf.reshape(m_dep, (-1, n_ls, 1, 1))
        cg_trans = tf.reshape(
            cg_trans, (n_ls, len(self.outs[0].spins), len(self.outs[1].spins))
        )
        H = tf.reduce_sum(m_dep * cg_trans, axis=1)
        # print(n_ls, cg_trans, self, m_dep.shape) # )data_p)
        if self.allow_cc:
            all_data = kwargs.get("all_data", {})
            charge = all_data.get("charge_conjugation", None)
            if charge is not None:
                H = tf.where(
                    charge[..., None, None] > 0, H, H[..., ::-1, ::-1]
                )
        ret = tf.reshape(
            H, (-1, 1, len(self.outs[0].spins), len(self.outs[1].spins))
        )
        return ret

    def get_angle_helicity_amp(self, data, data_p, **kwargs):
        m_dep = self.get_angle_ls_amp(data, data_p, **kwargs)
        cg_trans = tf.cast(self.get_cg_matrix(), m_dep.dtype)
        n_ls = len(self.get_ls_list())
        m_dep = tf.reshape(m_dep, (-1, n_ls, 1, 1))
        cg_trans = tf.reshape(
            cg_trans, (n_ls, len(self.outs[0].spins), len(self.outs[1].spins))
        )
        H = tf.reduce_sum(m_dep * cg_trans, axis=1)
        # print(n_ls, cg_trans, self, m_dep.shape) # )data_p)
        if self.allow_cc:
            all_data = kwargs.get("all_data", {})
            charge = all_data.get("charge_conjugation", None)
            if charge is not None:
                H = tf.where(
                    charge[..., None, None] > 0, H, H[..., ::-1, ::-1]
                )
        ret = tf.reshape(
            H, (-1, 1, len(self.outs[0].spins), len(self.outs[1].spins))
        )
        return ret

    def get_g_ls(self):
        gls = self.g_ls()
        if self.ls_index is None:
            return tf.stack(gls)
        # print(self, gls, self.ls_index)
        return tf.stack([gls[k] for k in self.ls_index])

    def get_ls_amp_org(self, data, data_p, **kwargs):
        g_ls = self.get_g_ls()
        # print(g_ls)
        q0 = self.get_relative_momentum(data_p, False)
        data["|q0|"] = q0
        if "|q|" in data:
            q = data["|q|"]
        else:
            q = self.get_relative_momentum(data_p, True)
            data["|q|"] = q
        if self.has_barrier_factor:
            bf = self.get_barrier_factor(data_p[self.core]["m"], q, q0, self.d)
            mag = g_ls
            m_dep = mag * tf.cast(bf, mag.dtype)
        else:
            m_dep = g_ls
        return m_dep

    def get_ls_amp(self, data, data_p, **kwargs):
        g_ls = self.get_g_ls()
        # print(g_ls)
        q0 = self.get_relative_momentum2(data_p, False)
        data["|q0|2"] = q0
        if "|q|2" in data:
            q = data["|q|2"]
        else:
            q = self.get_relative_momentum2(data_p, True)
            data["|q|2"] = q
        if self.has_barrier_factor:
            bf = self.get_barrier_factor2(
                data_p[self.core]["m"], q, q0, self.d
            )
            mag = g_ls
            m_dep = mag * tf.cast(bf, mag.dtype)
        else:
            m_dep = g_ls
        return m_dep

    def get_angle_g_ls(self):
        gls = [complex(1.0, 0.0) for i in self.g_ls()]
        if self.ls_index is None:
            return tf.stack(gls)
        return tf.stack([gls[k] for k in self.ls_index])

    def get_angle_ls_amp(self, data, data_p, **kwargs):
        g_ls = self.get_angle_g_ls()
        return g_ls

    def get_barrier_factor(self, mass, q, q0, d):
        ls = self.get_l_list()
        ret = []
        for l in ls:
            if self.has_bprime:
                tmp = q ** l * tf.cast(Bprime(l, q, q0, d), dtype=q.dtype)
            else:
                tmp = q ** l
            # tmp = tf.where(q > 0, tmp, tf.zeros_like(tmp))
            ret.append(tf.reshape(tmp, (-1, 1)))
        ret = tf.concat(ret, axis=-1)
        mass_dep = self.get_barrier_factor_mass(mass)
        return ret * mass_dep

    def get_barrier_factor2(self, mass, q2, q02, d):
        ls = self.get_l_list()
        ret = []
        for l in ls:
            if self.has_bprime:
                bp = Bprime_q2(l, q2, q02, d)
                tmp = q2 ** (l / 2) * tf.cast(bp, dtype=q2.dtype)
                if self.barrier_factor_norm:
                    tmp = tmp / q02 ** (l / 2)
            else:
                tmp = q2 ** (l / 2)
            # tmp = tf.where(q > 0, tmp, tf.zeros_like(tmp))
            ret.append(tf.reshape(tmp, (-1, 1)))
        ret = tf.concat(ret, axis=-1)
        mass_dep = self.get_barrier_factor_mass(mass)
        return ret * mass_dep

    def get_barrier_factor_mass(self, mass):
        if not self.barrier_factor_mass:
            return 1.0
        ls = tf.convert_to_tensor(self.get_l_list(), dtype=mass.dtype)
        m_dep = 1.0 / tf.pow(tf.expand_dims(mass, -1), ls)
        return m_dep

    def get_amp(self, data, data_p, **kwargs):
        a = self.core
        b = self.outs[0]
        c = self.outs[1]
        ang = data[b]["ang"]
        D_conj = get_D_matrix_lambda(ang, a.J, a.spins, b.spins, c.spins)
        H = self.get_helicity_amp(data, data_p, **kwargs)
        H = tf.reshape(
            H, (-1, 1, len(self.outs[0].spins), len(self.outs[1].spins))
        )
        H = tf.cast(H, dtype=D_conj.dtype)
        ret = H * tf.stop_gradient(D_conj)
        # print(self, H, D_conj)
        # exit()
        if self.aligned:
            for j, particle in enumerate(self.outs):
                if particle.J != 0 and "aligned_angle" in data[particle]:
                    ang = data[particle].get("aligned_angle", None)
                    if ang is None:
                        continue
                    dt = get_D_matrix_lambda(
                        ang, particle.J, particle.spins, particle.spins
                    )
                    dt_shape = [-1, 1, 1, 1, 1]
                    dt_shape[j + 2] = len(particle.spins)
                    dt_shape[j + 3] = len(particle.spins)
                    dt = tf.reshape(dt, dt_shape)
                    D_shape = [-1, len(a.spins), len(b.spins), len(c.spins)]
                    D_shape.insert(j + 3, 2)
                    D_shape[j + 3] = 1
                    ret = tf.reshape(ret, D_shape)
                    ret = dt * ret
                    ret = tf.reduce_sum(ret, axis=j + 2)
        return ret

    def get_angle_amp(self, data, data_p, **kwargs):
        a = self.core
        b = self.outs[0]
        c = self.outs[1]
        ang = data[b]["ang"]
        D_conj = get_D_matrix_lambda(ang, a.J, a.spins, b.spins, c.spins)
        H = self.get_angle_helicity_amp(data, data_p, **kwargs)
        H = tf.reshape(
            H, (-1, 1, len(self.outs[0].spins), len(self.outs[1].spins))
        )
        H = tf.cast(H, dtype=D_conj.dtype)
        ret = H * tf.stop_gradient(D_conj)
        # print(self, H, D_conj)
        # exit()
        if self.aligned:
            for j, particle in enumerate(self.outs):
                if particle.J != 0 and "aligned_angle" in data[particle]:
                    ang = data[particle].get("aligned_angle", None)
                    if ang is None:
                        continue
                    dt = get_D_matrix_lambda(
                        ang, particle.J, particle.spins, particle.spins
                    )
                    dt_shape = [-1, 1, 1, 1, 1]
                    dt_shape[j + 2] = len(particle.spins)
                    dt_shape[j + 3] = len(particle.spins)
                    dt = tf.reshape(dt, dt_shape)
                    D_shape = [-1, len(a.spins), len(b.spins), len(c.spins)]
                    D_shape.insert(j + 3, 2)
                    D_shape[j + 3] = 1
                    ret = tf.reshape(ret, D_shape)
                    ret = dt * ret
                    ret = tf.reduce_sum(ret, axis=j + 2)
        return ret

    def get_m_dep(self, data, data_p, **kwargs):
        return self.get_ls_amp(data, data_p, **kwargs)

    def get_ls_list(self):
        """get possible ls for decay, with l_list filter possible l"""
        ls_list = super(HelicityDecay, self).get_ls_list()
        if self.ls_list is not None:
            return self.ls_list
        if self.l_list is None:
            return ls_list
        ret = []
        for l, s in ls_list:
            if l in self.l_list:
                ret.append((l, s))
        return tuple(ret)


@regist_decay("default", 3)
@regist_decay("AngSam3", 3)
class AngSam3Decay(AmpDecay, AmpBase):
    def init_params(self):
        a = self.core.J
        self.gi = self.add_var(
            "G_mu", is_complex=True, shape=(_spin_int(2 * a + 1),)
        )
        try:
            self.gi.set_fix_idx(fix_idx=0, fix_vals=(1.0, 0.0))
        except Exception as e:
            print(e)

    def get_amp(self, data, data_extra=None, **kwargs):
        a = self.core
        b = self.outs[0]
        c = self.outs[1]
        d = self.outs[2]
        gi = tf.stack(self.gi())
        ang = data["ang"]
        D_conj = get_D_matrix_lambda(
            ang, a.J, a.spins, tuple(_spin_range(-a.J, a.J))
        )
        ret = tf.cast(gi, D_conj.dtype) * D_conj
        ret = tf.reduce_sum(ret, axis=-1)
        ret = tf.reshape(ret, (-1, len(a.spins), 1, 1, 1))
        ret = tf.tile(ret, [1, 1, len(b.spins), len(c.spins), len(d.spins)])
        return ret


class AmpDecayChain(BaseDecayChain, AmpBase):
    def __init__(self, *args, is_cp=False, **kwargs):
        self.is_cp = is_cp
        super(AmpDecayChain, self).__init__(*args, **kwargs)
        self.aligned = True
        self.need_amp_particle = True


@register_decay_chain("default")
class DecayChain(AmpDecayChain):
    """A list of Decay as a chain decay"""

    def init_params(self, name=""):
        self.total = self.add_var(
            name + "total", is_complex=True, is_cp=self.is_cp, shape=[1]
        )
        # self.total = self.add_var(name + "total", is_complex=True, shape=[1])

    def get_factor_variable(self):
        a = []
        for i in self:
            tmp = i.get_factor_variable()
            if tmp:
                a.append(tmp)
        for j in self.inner:
            tmp = j.get_factor_variable()
            if tmp:
                a.append(tmp)
        return [tuple([self.total] + a)]

    def get_amp_total(self, charge=1):
        return tf.stack(self.total(charge))

    def product_gls(self):
        ret = self.get_all_factor()
        return tf.reduce_prod(ret)

    def get_all_factor(self):
        ret = [self.get_amp_total()]
        for i in self:
            ret.append(i.get_g_ls())
        return ret

    def get_cp_amp_total(self, charge=1):
        if not self.is_cp:
            return self.get_amp_total()
        total_pos = self.get_amp_total(1)
        total_neg = self.get_amp_total(-1)
        # print("total_pos", total_pos)
        # print("total_neg", total_neg)
        charge_cond = charge > 0
        # print("charge", charge)
        total = tf.where(charge_cond, total_pos, total_neg)
        return total

    def get_amp(self, data_c, data_p, all_data=None, base_map=None):
        base_map = self.get_base_map(base_map)
        iter_idx = ["..."]
        amp_d = []
        indices = []
        final_indices = "".join(iter_idx + self.amp_index(base_map))
        for i in self:
            indices.append(i.amp_index(base_map))
            amp_d.append(i.get_amp(data_c[i], data_p, all_data=all_data))

        if self.need_amp_particle:
            rs = self.get_amp_particle(data_p, data_c, all_data=all_data)

            total = self.get_cp_amp_total(
                charge=all_data.get("charge_conjugation", 1)
            )
            if rs is not None:
                total = total * tf.cast(rs, total.dtype)
            # print(total)*self.get_amp_total()
            amp_d.append(total)
            indices.append([])

        if self.aligned:
            for i in self:
                for j in i.outs:
                    if j.J != 0 and "aligned_angle" in data_c[i][j]:
                        ang = data_c[i][j]["aligned_angle"]
                        dt = get_D_matrix_lambda(ang, j.J, j.spins, j.spins)
                        amp_d.append(tf.stop_gradient(dt))
                        idx = [base_map[j], base_map[j].upper()]
                        indices.append(idx)
                        final_indices = final_indices.replace(*idx)
        idxs = []
        for i in indices:
            tmp = "".join(iter_idx + i)
            idxs.append(tmp)
        idx = ",".join(idxs)
        idx_s = "{}->{}".format(idx, final_indices)
        # ret = amp * tf.reshape(rs, [-1] + [1] * len(self.amp_shape()))
        # print(idx_s)#, amp_d)
        try:
            ret = einsum(idx_s, *amp_d)
        except:
            ret = tf.einsum(idx_s, *amp_d)
        # print(self, ret[0])
        # exit()
        # ret = einsum(idx_s, *amp_d)
        return ret

    def get_angle_amp(self, data_c, data_p, all_data=None, base_map=None):
        base_map = self.get_base_map(base_map)
        iter_idx = ["..."]
        amp_d = []
        indices = []
        final_indices = "".join(iter_idx + self.amp_index(base_map))
        for i in self:
            indices.append(i.amp_index(base_map))
            amp_d.append(i.get_angle_amp(data_c[i], data_p, all_data=all_data))

        if self.aligned:
            for i in self:
                for j in i.outs:
                    if j.J != 0 and "aligned_angle" in data_c[i][j]:
                        ang = data_c[i][j]["aligned_angle"]
                        dt = get_D_matrix_lambda(ang, j.J, j.spins, j.spins)
                        amp_d.append(tf.stop_gradient(dt))
                        idx = [base_map[j], base_map[j].upper()]
                        indices.append(idx)
                        final_indices = final_indices.replace(*idx)
        idxs = []
        for i in indices:
            tmp = "".join(iter_idx + i)
            idxs.append(tmp)
        idx = ",".join(idxs)
        idx_s = "{}->{}".format(idx, final_indices)
        # ret = amp * tf.reshape(rs, [-1] + [1] * len(self.amp_shape()))
        # print(idx_s)#, amp_d)
        try:
            ret = einsum(idx_s, *amp_d)
        except:
            ret = tf.einsum(idx_s, *amp_d)
        # print(self, ret[0])
        # exit()
        # ret = einsum(idx_s, *amp_d)
        return ret

    def get_m_dep(self, data_c, data_p, all_data=None, base_map=None):
        base_map = self.get_base_map(base_map)
        iter_idx = ["..."]
        amp_d = []
        indices = []
        final_indices = "".join(iter_idx + self.amp_index(base_map))
        for i in self:
            indices.append(i.amp_index(base_map))
            amp_d.append(i.get_m_dep(data_c[i], data_p, all_data=all_data))

        if self.need_amp_particle:
            rs = self.get_amp_particle(data_p, data_c, all_data=all_data)
            total = self.get_cp_amp_total(
                all_data.get("charge_conjugation", 1)
            )
            # print("total_pos", total_pos)
            # print("total_neg", total_neg)
            if rs is not None:
                total = total * tf.cast(rs, total.dtype)
                # print("charge", charge)
            # print(total)
            # print(total)*self.get_amp_total()
            amp_d.append(total)
        return amp_d

    def get_amp_particle(self, data_p, data_c, all_data=None):
        amp_p = []
        if not self.inner:
            return 1.0
        for i in self.inner:
            if len(i.decay) >= 1:
                decay_i = i.decay[0]
                found = False
                for j in i.decay:
                    if j in self:
                        decay_i = j
                        found = True
                        break
                if not found:
                    raise IndexError(
                        "not found {} decay in {}".format(i, self)
                    )
                data_c_i = data_c[decay_i]
                if "|q|" not in data_c_i:
                    data_c_i["|q|"] = decay_i.get_relative_momentum(
                        data_p, True
                    )
                if "|q0|" not in data_c_i:
                    data_c_i["|q0|"] = decay_i.get_relative_momentum(
                        data_p, False
                    )
                if "|q|2" not in data_c_i:
                    data_c_i["|q|2"] = decay_i.get_relative_momentum2(
                        data_p, True
                    )
                if "|q0|2" not in data_c_i:
                    data_c_i["|q0|2"] = decay_i.get_relative_momentum2(
                        data_p, False
                    )
                amp_p.append(i.get_amp(data_p[i], data_c_i, all_data=all_data))
            else:
                amp_p.append(i.get_amp(data_p[i], all_data=all_data))
        rs = 1.0
        for i in amp_p:
            rs = rs * i
        # tf.reduce_prod(amp_p, axis=0)
        return rs

    def amp_shape(self):
        ret = [len(self.top.spins)]
        for i in self.outs:
            ret.append(len(i.spins))
        return tuple(ret)

    # @simple_cache_fun
    def amp_index(self, base_map=None):
        if base_map is None:
            base_map = self.get_base_map()
        ret = [base_map[self.top]]
        for i in self.outs:
            ret.append(base_map[i])
        return ret

    def get_base_map(self, base_map=None):
        gen = index_generator(base_map)
        if base_map is None:
            base_map = {}
        ret = base_map.copy()
        if self.top not in base_map:
            ret[self.top] = next(gen)
        for i in self.outs:
            if i not in base_map:
                ret[i] = next(gen)
        for i in self.inner:
            if i not in ret:
                ret[i] = next(gen)
        return ret


class DecayGroup(BaseDecayGroup, AmpBase):
    """ A Group of Decay Chains with the same final particles."""

    def __init__(self, chains):
        self.chains_idx = list(range(len(chains)))
        first_chain = chains[0]
        if not isinstance(first_chain, DecayChain):
            chains = [DecayChain(i) for i in chains]
        super(DecayGroup, self).__init__(chains)
        self.not_full = False
        self.polarization = getattr(self.top, "polarization", "none")
        # self.init_params()

    def init_params(self, name=""):
        for i in self.resonances:
            i.init_params()
        inited_set = set()
        for i in self:
            i.init_params(name)
            for j in i:
                if j not in inited_set:
                    j.init_params()
                    inited_set.add(j)
        if self.polarization == "vector":
            print("add polarization vector")
            if self.top.J == 0.5:
                self.polarization_vector = [
                    self.top.add_var("polarization_px"),
                    self.top.add_var("polarization_py"),
                    self.top.add_var("polarization_pz"),
                ]

    def get_factor_variable(self):
        ret = []
        for i in self:
            ret += i.get_factor_variable()
        return ret

    def get_amp(self, data):
        """
        calculate the amplitude as complex number
        """
        data_particle = data["particle"]
        data_decay = data["decay"]

        used_chains = tuple([self.chains[i] for i in self.chains_idx])
        chain_maps = self.get_chains_map(used_chains)
        base_map = self.get_base_map()
        ret = []
        for chains in chain_maps:
            for decay_chain in chains:
                chain_topo = decay_chain.standard_topology()
                found = False
                for i in data_decay.keys():
                    if i == chain_topo:
                        data_decay_i = data_decay[i]
                        found = True
                        break
                if not found:
                    raise KeyError("not found {}".format(chain_topo))
                data_c = rename_data_dict(data_decay_i, chains[decay_chain])
                data_p = rename_data_dict(data_particle, chains[decay_chain])
                # print("$$$$$",data_c)
                # print("$$$$$",data_p)
                amp = decay_chain.get_amp(
                    data_c, data_p, base_map=base_map, all_data=data
                )
                ret.append(amp)
                # print(decay_chain, amp[:10])
        ret = tf.reduce_sum(ret, axis=0)
        return ret

    def get_m_dep(self, data):
        """get mass dependent items"""
        data_particle = data["particle"]
        data_decay = data["decay"]

        used_chains = tuple([self.chains[i] for i in self.chains_idx])
        chain_maps = self.get_chains_map(used_chains)
        base_map = self.get_base_map()
        ret = []
        for decay_chain in used_chains:
            for chains in chain_maps:
                if str(decay_chain) in [str(i) for i in chains]:
                    maps = chains[decay_chain]
                    break
            chain_topo = decay_chain.standard_topology()
            found = False
            for i in data_decay.keys():
                if i == chain_topo:
                    data_decay_i = data_decay[i]
                    found = True
                    break
            if not found:
                raise KeyError("not found {}".format(chain_topo))
            data_c = rename_data_dict(data_decay_i, maps)
            data_p = rename_data_dict(data_particle, maps)
            # print("$$$$$",data_c)
            # print("$$$$$",data_p)
            amp = decay_chain.get_m_dep(
                data_c, data_p, base_map=base_map, all_data=data
            )
            ret.append(amp)
        # ret = tf.reduce_sum(ret, axis=0)
        return ret

    def get_angle_amp(self, data):
        data_particle = data["particle"]
        data_decay = data["decay"]

        used_chains = tuple([self.chains[i] for i in self.chains_idx])
        chain_maps = self.get_chains_map(used_chains)
        base_map = self.get_base_map()
        ret = []
        for decay_chain in used_chains:
            for chains in chain_maps:
                if str(decay_chain) in [str(i) for i in chains]:
                    maps = chains[decay_chain]
                    break
            chain_topo = decay_chain.standard_topology()
            found = False
            for i in data_decay.keys():
                if i == chain_topo:
                    data_decay_i = data_decay[i]
                    found = True
                    break
            if not found:
                raise KeyError("not found {}".format(chain_topo))
            data_c = rename_data_dict(data_decay_i, maps)
            data_p = rename_data_dict(data_particle, maps)
            amp = decay_chain.get_angle_amp(
                data_c, data_p, base_map=base_map, all_data=data
            )
            ret.append(amp)
        # ret = tf.reduce_sum(ret, axis=0)
        return amp

    @functools.lru_cache()
    def get_swap_factor(self, key):
        factor = 1.0
        for i, j in zip(self.identical_particles, key[1]):
            p = self.get_particle(i[0])
            if int(p.J * 2) % 2 == 0:
                continue
            for m, n in zip(i, j):
                if m != n:
                    factor *= -1.0
        return factor

    def get_amp2(self, data):
        amp = self.get_amp(data)
        id_swap = data.get("id_swap", {})
        for k, v in id_swap.items():
            new_data = {**data, **v}
            factor = self.get_swap_factor(k)
            amp_swap = factor * self.get_amp(new_data)
            # print(k, amp, amp_swap)
            amp = amp + amp_swap
        return amp

    def sum_amp(self, data, cached=True):
        """
        calculat the amplitude modular square
        """
        if not cached:
            data = simple_deepcopy(data)
        if self.polarization != "none":
            return self.sum_amp_polarization(data)
        amp = self.get_amp2(data)
        amp2s = tf.math.real(amp * tf.math.conj(amp))
        idx = list(range(1, len(amp2s.shape)))
        sum_A = tf.reduce_sum(amp2s, idx)
        return sum_A

    def sum_amp_polarization(self, data):
        """
        sum amplitude suqare with density _get_cg_matrix

        .. math::
            P = \\sum_{m, m', \\cdots } A_{m, \\cdots}  \\rho_{m, m'} A^{*}_{m', \\cdots}

        """

        amp = self.get_amp(data)
        amp = tf.reshape(
            amp, (amp.shape[0], amp.shape[1], -1)
        )  # (i, la, lb lc ld ...)
        na, nl = amp.shape[1], amp.shape[2]
        rho = self.get_density_matrix()
        amp = tf.reshape(amp, (-1, na, 1, nl))
        amp_c = tf.reshape(
            tf.math.conj(amp), (-1, na, nl)
        )  # (i, la, lb lc ld ...)
        sum_A = (
            tf.reduce_sum(amp * tf.reshape(rho, (na, na, 1)), axis=1) * amp_c
        )
        return tf.reduce_sum(tf.math.real(sum_A), axis=[1, 2])

    def get_density_matrix(self):
        if self.polarization == "vector":
            px, py, pz = [i() for i in self.polarization_vector]
            zeros = tf.zeros_like(px)
            ones = tf.ones_like(px)
            rho00 = tf.complex(ones + pz, zeros)
            rho11 = tf.complex(ones - pz, zeros)
            rho01 = tf.complex(px, -py)
            rho10 = tf.complex(px, py)
            ret = 0.5 * tf.stack([[rho00, rho01], [rho10, rho11]])
            # print(ret)
            return ret
        raise NotImplementedError

    # @simple_cache_fun
    def amp_index(self, gen=None, base_map=None):
        if base_map is None:
            base_map = self.get_base_map()
        ret = [base_map[self.top]]
        for i in self.outs:
            ret.append(base_map[i])
        return ret

    def get_base_map(self, gen=None, base_map=None):
        if gen is None:
            gen = index_generator(base_map)
        if base_map is None:
            base_map = {self.top: next(gen)}
        for i in self.outs:
            base_map[i] = next(gen)
        return base_map

    def get_res_map(self):
        res_map = {}
        for i, decay in enumerate(self.chains):
            for j in decay.inner:
                if j not in res_map:
                    res_map[j] = []
                res_map[j].append(i)
        return res_map

    def set_used_res(self, res, only=False):
        res_set = set()
        for i in res:
            if isinstance(i, str):
                res_set.add(BaseParticle(i))
            elif isinstance(i, BaseParticle):
                res_set.add(i)
            else:
                raise TypeError(
                    "type({}) = {} not a Particle".format(i, type(i))
                )
        if not only:
            used_res = set()
            for i in res_set:
                for j, c in enumerate(self.chains):
                    if i in c.inner:
                        used_res.add(j)
            self.set_used_chains(list(used_res))
        else:
            unused_res = set(self.resonances) - res_set
            unused_decay = set()
            res_map = self.get_res_map()
            for i in unused_res:
                for j in res_map[i]:
                    unused_decay.add(j)
            used_decay = []
            for i, _ in enumerate(self.chains):
                if i not in unused_decay:
                    used_decay.append(i)
            self.set_used_chains(used_decay)

    def set_used_chains(self, used_chains):
        self.chains_idx = list(used_chains)
        if len(self.chains_idx) != len(self.chains):
            self.not_full = True

    def partial_weight(self, data, combine=None):
        chains = list(self.chains)
        if combine is None:
            combine = [[i] for i in range(len(chains))]
        o_used_chains = self.chains_idx
        weights = []
        for i in combine:
            self.set_used_chains(i)
            weight = self.sum_amp(data)
            weights.append(weight)
        self.set_used_chains(o_used_chains)
        return weights

    def chains_particle(self):
        ret = []
        for i in self:
            ret.append(tuple(i.inner))
        return ret

    def partial_weight_interference(self, data):
        chains = list(self.chains)
        combine = combinations(range(len(chains)), 2)
        o_used_chains = self.chains_idx
        weights = {}
        for i in combine:
            self.set_used_chains(i)
            weight = self.sum_amp(data)
            weights[i] = weight
        self.set_used_chains(o_used_chains)
        return weights

    def generate_phasespace(self, num=100000):
        def get_mass(i):
            mass = i.get_mass()
            if mass is None:
                raise Exception("mass is required for particle {}".format(i))
            return mass

        top_mass = get_mass(self.top)
        final_mass = [get_mass(i) for i in self.outs]
        from tf_pwa.phasespace import PhaseSpaceGenerator

        a = PhaseSpaceGenerator(top_mass, final_mass)
        data = a.generate(num)
        return dict(zip(self.outs, data))


def index_generator(base_map=None):
    indices = "abcdefghjklmnopqrstuvwxyz"
    if base_map is not None:
        for i in base_map:
            indices = indices.replace(base_map[i], "")
    for i in indices:
        yield i


def rename_data_dict(data, idx_map):
    if isinstance(data, dict):
        return {
            idx_map.get(k, k): rename_data_dict(v, idx_map)
            for k, v in data.items()
        }
    if isinstance(data, tuple):
        return tuple([rename_data_dict(i, idx_map) for i in data])
    if isinstance(data, list):
        return [rename_data_dict(i, idx_map) for i in data]
    return data


def value_and_grad(f, var):
    with tf.GradientTape() as tape:
        s = f(var)
    g = tape.gradient(s, var)
    return s, g


class AmplitudeModel(object):
    def __init__(
        self, decay_group, name="", polar=None, vm=None, use_tf_function=False
    ):
        self.decay_group = decay_group
        self._name = name
        with variable_scope(vm) as vm:
            if polar is not None:
                vm.polar = polar
            decay_group.init_params(name)
        self.vm = vm
        res = decay_group.resonances
        self.used_res = res
        self.res = res
        self.f_data = []
        if use_tf_function:
            self.cached_fun = tf.function(
                self.decay_group.sum_amp, experimental_relax_shapes=True
            )
        else:
            self.cached_fun = self.decay_group.sum_amp

    def __del__(self):
        if hasattr(self, "cached_fun"):
            del self.cached_fun
        # super(AmplitudeModel, self).__del__()

    def cache_data(self, data, split=None, batch=None):
        for i in self.decay_group:
            for j in i.inner:
                print(j)
        if split is None and batch is None:
            return data
        else:
            n = data_shape(data)
            if batch is None:  # split个一组，共batch组
                batch = (n + split - 1) // split
            ret = list(split_generator(data, batch))
            return ret

    def set_used_res(self, res):
        self.decay_group.set_used_res(res)

    def set_used_chains(self, used_chains):
        self.decay_group.set_used_chains(used_chains)

    def partial_weight(self, data, combine=None):
        return self.decay_group.partial_weight(data, combine)

    def partial_weight_interference(self, data):
        return self.decay_group.partial_weight_interference(data)

    def get_params(self, trainable_only=False):
        return self.vm.get_all_dic(trainable_only)

    def set_params(self, var):
        self.vm.set_all(var)

    @contextlib.contextmanager
    def temp_params(self, var):
        params = self.get_params()
        self.set_params(var)
        yield var
        self.set_params(params)

    def chains_particle(self):
        return self.decay_group.chains_particle()

    @property
    def variables(self):
        return self.vm.variables

    @property
    def trainable_variables(self):
        return self.vm.trainable_variables

    def __call__(self, data, cached=False):
        if id(data) in self.f_data:
            if not self.decay_group.not_full:
                return self.cached_fun(data)
        else:
            self.f_data.append(id(data))
        ret = self.decay_group.sum_amp(data)
        return ret


def load_decfile_particle(fname):
    with open(fname) as f:
        dec = load_dec_file(f)
    dec = list(dec)
    particles = {}

    def get_particles(name):
        if name not in particles:
            a = get_particle(name)
            particles[name] = a
        return particles[name]

    decay = []
    for i in dec:
        cmd, var = i
        if cmd == "Particle":
            a = get_particles(var["name"])
            setattr(a, "params", var["params"])
        if cmd == "Decay":
            for j in var["final"]:
                outs = [get_particles(k) for k in j["outs"]]
                de = Decay(get_particles(var["name"]), outs)
                for k in j:
                    if k != "outs":
                        setattr(de, k, j[k])
                decay.append(de)
        if cmd == "RUNNINGWIDTH":
            pa = get_particles(var[0])
            setattr(pa, "running_width", True)
    top, inner, outs = split_particle_type(decay)
    return top, inner, outs


regist_config(DEFAULT_DECAY, (HelicityDecay, {}))



import numpy as np
import pytest

from tf_pwa.phasespace import *


def test_phasespace():
    a = PhaseSpaceGenerator(10, [3, 2, 1])
    data = a.generate(100)
    assert len(data), 3
    for i in data:
        assert i.shape == (100, 4)

    assert np.allclose(LorentzVector.M(data[0]), [3] * 100)
    assert np.allclose(LorentzVector.M(data[1]), [2] * 100)
    assert np.allclose(LorentzVector.M(data[2]), [1] * 100)
    p_all = data[0] + data[1] + data[2]
    assert np.allclose(LorentzVector.M(p_all), [10] * 100)
    assert np.allclose(p_all[:, 1:3], np.zeros_like(p_all[:, 1:3]))


def test_generate_phsp():
    (a, b), c = generate_phsp(5.0, ((3.0, (1.0, 1.0)), 1.0))
    assert np.allclose(LorentzVector.M(a + b + c), 5.0)
    # assert np.sum(np.abs((a+b+c)[:,1:])) < 2e-5


def test_error():
    with pytest.raises(ValueError):
        PhaseSpaceGenerator(1, [0.3, 0.4, 0.5]).generate(10)



import numpy as np

from tf_pwa.amp.core import Particle, register_particle
from tf_pwa.tensorflow_wrapper import tf

# pylint: disable=no-member


class InterpolationParticle(Particle):
    def __init__(self, *args, **kwargs):
        self.points = None
        self.max_m = None
        self.min_m = None
        self.interp_N = None
        self.polar = True
        self.fix_idx = -1
        self.with_bound = False
        super(InterpolationParticle, self).__init__(*args, **kwargs)
        self.fix_width = True
        if self.points is None:
            dx = (self.max_m - self.min_m) / (self.interp_N - 1)
            self.points = [self.min_m + dx * i for i in range(self.interp_N)]
        else:
            self.fix_width = False
            self.interp_N = len(self.points)
        self.bound = [
            (self.points[i], self.points[i + 1])
            for i in range(0, self.interp_N - 1)
        ]
        if self.fix_idx is not None and self.fix_idx < 0:
            self.fix_idx = self.interp_N // 2 - 1

    def init_params(self):
        # self.a = self.add_var("a")
        self.point_value = self.add_var(
            "point",
            is_complex=True,
            shape=(self.n_points(),),
            polar=self.polar,
        )
        if self.fix_idx is not None:
            self.point_value.set_fix_idx(fix_idx=self.fix_idx, fix_vals=1.0)

    def get_amp(self, data, *args, **kwargs):
        m = data["m"]
        fm = self.interp(m)
        return fm

    def n_points(self):
        if self.with_bound:
            return self.interp_N
        return self.interp_N - 2

    def __call__(self, mass):
        return self.interp(mass)

    def interp(self, mass):
        raise NotImplementedError

    def get_point_values(self):
        p = self.point_value()
        v_r = [0.0] + [tf.math.real(i) for i in p] + [0.0]
        v_i = [0.0] + [tf.math.imag(i) for i in p] + [0.0]
        return self.points, v_r, v_i

    def get_bin_index(self, m):
        if self.fix_width:
            m_min = tf.convert_to_tensor(self.points[0], m.dtype)
            m_max = tf.convert_to_tensor(self.points[-1], m.dtype)
            delta_width = (m_max - m_min) / (self.interp_N - 1)
            bin_idx = tf.histogram_fixed_width_bins(
                m,
                [m_min - delta_width, m_max + delta_width],
                nbins=self.interp_N + 1,
                dtype=tf.dtypes.int64,
            )
        else:
            # dig = lambda x, y: tf.numpy_function(np.digitize, [x, y], tf.int64)
            # bin_idx = dig(m, self.points)
            bin_idx = tf.raw_ops.Bucketize(input=m, boundaries=self.points)
        bin_idx = bin_idx - 1
        # print(tf.reduce_max(bin_idx), tf.reduce_min(bin_idx))
        bin_idx = tf.stop_gradient(bin_idx)
        return bin_idx


@register_particle("interp")
class Interp(InterpolationParticle):
    """linear interpolation for real number"""

    def init_params(self):
        # self.a = self.add_var("a")
        self.point_value = self.add_var("point", shape=(self.interp_N + 1,))
        self.point_value.set_fix_idx(fix_idx=self.fix_idx, fix_vals=1.0)

    def interp(self, m):
        # q = data_extra[self.outs[0]]["|q|"]
        # a = self.a()
        zeros = tf.zeros_like(m)
        p = tf.abs(self.point_value())

        def add_f(x, bl, br, pl, pr):
            return tf.where(
                (x > bl) & (x <= br),
                (x - bl) / (br - bl) * (pr - pl) + pl,
                zeros,
            )

        ret = [
            add_f(m, self.points[i], self.points[i + 1], p[i], p[i + 1])
            for i in range(self.interp_N - 1)
        ]
        return tf.complex(tf.reduce_sum(ret, axis=0), zeros)


@register_particle("interp_c")
class Interp(InterpolationParticle):
    """linear interpolation for complex number"""

    def interp(self, m):
        # q = data_extra[self.outs[0]]["|q|"]
        # a = self.a()
        p = self.point_value()
        zeros = tf.zeros_like(m)
        ones = tf.ones_like(m)

        def poly_i(i, xi):
            tmp = zeros
            for j in range(i - 1, i + 1):
                if j < 0 or j > self.interp_N - 1:
                    continue
                r = ones
                for k in range(j, j + 2):
                    if k == i:
                        continue
                    r = r * (m - xi[k]) / (xi[i] - xi[k])
                r = tf.where((m >= xi[j]) & (m < xi[j + 1]), r, zeros)
                tmp = tmp + r
            return tmp

        h = tf.stack(
            [poly_i(i, self.points) for i in range(1, self.interp_N - 1)],
            axis=-1,
        )
        h = tf.stop_gradient(h)
        p_r = tf.math.real(p)
        p_i = tf.math.imag(p)
        ret_r = tf.reduce_sum(h * p_r, axis=-1)
        ret_i = tf.reduce_sum(h * p_i, axis=-1)
        return tf.complex(ret_r, ret_i)


@register_particle("spline_c")
class Interp1DSpline(InterpolationParticle):
    """Spline interpolation function for model independent resonance"""

    def __init__(self, *args, **kwargs):
        self.bc_type = "not-a-knot"
        super(Interp1DSpline, self).__init__(*args, **kwargs)
        assert self.interp_N > 2, "points need large than 2"
        self.h_matrix = None

    def init_params(self):
        super(Interp1DSpline, self).init_params()
        h_matrix = spline_xi_matrix(self.points, self.bc_type)
        if self.with_bound:
            self.h_matrix = tf.convert_to_tensor(h_matrix)
        else:
            self.h_matrix = tf.convert_to_tensor(h_matrix[..., 1:-1])

    def interp(self, m):
        zeros = tf.zeros_like(m)
        p = self.point_value()
        p_r = tf.math.real(p)
        p_i = tf.math.imag(p)
        xi_m = self.h_matrix
        x_m = spline_x_matrix(m, self.points)
        x_m = tf.expand_dims(x_m, axis=-1)
        m_xi = tf.reduce_sum(xi_m * x_m, axis=[-3, -2])
        m_xi = tf.stop_gradient(m_xi)
        ret_r = tf.reduce_sum(tf.cast(m_xi, p_r.dtype) * p_r, axis=-1)
        ret_i = tf.reduce_sum(tf.cast(m_xi, p_i.dtype) * p_i, axis=-1)
        return tf.complex(ret_r, ret_i)


def spline_x_matrix(x, xi):
    """build matrix of x for spline interpolation"""
    ones = tf.ones_like(x)
    x2 = x * x
    x3 = x2 * x
    x_p = tf.stack([ones, x, x2, x3], axis=-1)
    x = tf.expand_dims(x, axis=-1)
    zeros = tf.zeros_like(x)

    def poly_i(i):
        cut = (x >= xi[i]) & (x < xi[i + 1])
        return tf.where(cut, x_p, zeros)

    xs = [poly_i(i) for i in range(len(xi) - 1)]
    return tf.stack(xs, axis=-2)


def spline_matrix(x, xi, yi, bc_type="not-a-knot"):
    """calculate spline interpolation"""
    xi_m = spline_xi_matrix(xi)  # (N_range, 4, N_yi)
    x_m = spline_x_matrix(x, xi)  # (..., N_range, 4)
    x_m = tf.expand_dims(x_m, axis=-1)
    m = tf.reduce_sum(xi_m * x_m, axis=[-3, -2])
    return tf.reduce_sum(tf.cast(m, yi.dtype) * yi, axis=-1)


def spline_xi_matrix(xi, bc_type="not-a-knot"):
    """build matrix of xi for spline interpolation
    solve equation

    .. math::
        S_i'(x_i) = S_{i-1}'(x_i)

    and two bound condition. :math:`S_0'(x_0) = S_{n-1}'(x_n) = 0`
    """
    N = len(xi)
    hi = [xi[i + 1] - xi[i] for i in range(N - 1)]

    h_matrix = np.zeros((N, N))
    if bc_type == "not-a-knot":
        h_matrix[0, 0] = -hi[1]
        h_matrix[0, 1] = hi[0] + hi[1]
        h_matrix[0, 2] = -hi[0]
    elif bc_type == "clamped":
        h_matrix[0, 0] = 2 * hi[0]
        h_matrix[0, 1] = hi[0]
    elif bc_type == "natural":
        h_matrix[0, 0] = 1
    else:
        raise ValueError("bc_type={} not in {not-a-knot,clamped,natural}")
    for i in range(1, N - 1):
        h_matrix[i, i - 1] = hi[i - 1]
        h_matrix[i, i] = 2 * (hi[i - 1] + hi[i])
        h_matrix[i, i + 1] = hi[i]
    if bc_type == "not-a-knot":
        h_matrix[-1, -3] = -hi[-1]
        h_matrix[-1, -2] = hi[-1] + hi[-2]
        h_matrix[-1, -1] = -hi[-2]
    elif bc_type == "clamped":
        h_matrix[-1, -2] = hi[-1]
        h_matrix[-1, -1] = 2 * hi[-1]
    elif bc_type == "natural":
        h_matrix[-1, -1] = 1
    h_matrix_inv = np.linalg.inv(h_matrix)
    y_matrix = np.zeros((N, N))
    if bc_type == "not-a-knot":
        y_matrix[0, 0] = 0  # 6 / hi[0]
    elif bc_type == "clamped":
        y_matrix[0, 0] = 6 / hi[0]
    elif bc_type == "natural":
        y_matrix[0, 0] = 0  # 6 / hi[0]
    for i in range(1, N - 1):
        y_matrix[i, i - 1] = 6 / hi[i - 1]
        y_matrix[i, i] = -6 * (1 / hi[i] + 1 / hi[i - 1])
        y_matrix[i, i + 1] = 6 / hi[i]
    if bc_type == "not-a-knot":
        y_matrix[-1, -1] = 0  # -6 / hi[-1]
    elif bc_type == "clamped":
        y_matrix[-1, -1] = -6 / hi[-1]
    elif bc_type == "natural":
        y_matrix[-1, -1] = 0  # -6 / hi[-1]

    hy_matrix = np.dot(h_matrix_inv, y_matrix)

    # Si(x) = ai + bi(x-xi) + ci(x-xi)^2 + di(x-xi)^3
    hi = np.array(hi)[:, np.newaxis]
    I = np.eye(N)
    ci = hy_matrix[:-1] / 2
    di = (hy_matrix[1:] - hy_matrix[:-1]) / 6 / hi
    bi = (I[1:] - I[:-1]) / hi - ci * hi - di * hi * hi
    ai = I[:-1]

    # Si(x) = ai + bi x + ci x^2 + di x^3
    x1 = np.array(xi[:-1])[:, np.newaxis]
    x2 = x1 * x1
    x3 = x2 * x1
    ai_2 = ai - bi * x1 + ci * x2 - di * x3
    bi_2 = bi - 2 * ci * x1 + 3 * di * x2
    ci_2 = ci - 3 * di * x1
    di_2 = di
    ret = np.stack([ai_2, bi_2, ci_2, di_2], axis=-2)
    return ret


@register_particle("interp1d3")
class Interp1D3(InterpolationParticle):
    """Piecewise third order interpolation"""

    def interp(self, m):
        p = self.point_value()
        ret = interp1d3(m, self.points, tf.stack(p))
        return ret


def interp1d3(x, xi, yi):
    h, b = get_matrix_interp1d3(x, xi)  # (..., N), (...,)
    ret = tf.reshape(
        tf.matmul(tf.cast(h, yi.dtype), tf.reshape(yi, (-1, 1))), b.shape
    ) + tf.cast(b, yi.dtype)
    return ret


def get_matrix_interp1d3(x, xi):
    N = len(xi) - 1
    zeros = tf.zeros_like(x)
    ones = tf.ones_like(x)
    # @pysnooper.snoop()
    def poly_i(i):
        tmp = zeros
        for j in range(i - 1, i + 3):
            if j < 0 or j > N - 1:
                continue
            r = ones
            for k in range(j - 1, j + 3):
                if k == i or k < 0 or k > N:
                    continue
                r = r * (x - xi[k]) / (xi[i] - xi[k])
            r = tf.where((x >= xi[j]) & (x < xi[j + 1]), r, zeros)
            tmp = tmp + r
        return tmp

    h = tf.stack([poly_i(i) for i in range(1, N)], axis=-1)
    b = tf.zeros_like(x)
    return h, b


@register_particle("interp_lagrange")
class Interp1DLang(InterpolationParticle):
    """Lagrange interpolation"""

    def interp(self, m):
        zeros = tf.zeros_like(m)
        p = self.point_value()
        xs = []

        def poly_i(i):
            x = 1.0
            for j in range(self.interp_N):
                if i == j:
                    continue
                x = (
                    x
                    * (m - self.points[j])
                    / (self.points[i] - self.points[j])
                )
            return x

        xs = tf.stack([poly_i(i) for i in range(self.interp_N)], axis=-1)
        zeros = tf.zeros_like(xs)
        xs = tf.complex(xs, zeros)
        ret = tf.reduce_sum(xs[:, 1:-1] * p, axis=-1)
        return ret


@register_particle("interp_hist")
class InterpHist(InterpolationParticle):
    """Interpolation for each bins as constant"""

    def interp(self, m):
        p = self.point_value()
        ones = tf.ones_like(m)
        zeros = tf.zeros_like(m)

        def add_f(x, bl, br):
            return tf.where((x > bl) & (x <= br), ones, zeros)

        x_bin = tf.stack(
            [
                add_f(
                    m,
                    (self.points[i] + self.points[i + 1]) / 2,
                    (self.points[i + 1] + self.points[i + 2]) / 2,
                )
                for i in range(self.interp_N - 2)
            ],
            axis=-1,
        )
        p_r = tf.math.real(p)
        p_i = tf.math.imag(p)
        x_bin = tf.stop_gradient(x_bin)
        ret_r = tf.reduce_sum(x_bin * p_r, axis=-1)
        ret_i = tf.reduce_sum(x_bin * p_i, axis=-1)
        return tf.complex(ret_r, ret_i)


class HistParticle(InterpolationParticle):
    def n_points(self):
        return self.interp_N - 1


@register_particle("hist_idx")
class InterpHistIdx(HistParticle):
    """Interpolation for each bins as constant"""

    def interp(self, m):
        _, p_r, p_i = self.get_point_values()
        bin_idx = self.get_bin_index(m)
        ret_r = tf.gather(p_r[1:], bin_idx)
        ret_i = tf.gather(p_i[1:], bin_idx)
        return tf.complex(ret_r, ret_i)


@register_particle("spline_c_idx")
class Interp1DSplineIdx(InterpolationParticle):
    """Spline function in index way"""

    def __init__(self, *args, **kwargs):
        self.bc_type = "not-a-knot"
        super().__init__(*args, **kwargs)
        assert self.interp_N > 2, "points need large than 2"
        self.h_matrix = None

    def init_params(self):
        super(Interp1DSplineIdx, self).init_params()
        h_matrix = spline_xi_matrix(self.points, self.bc_type)
        if self.with_bound:
            self.h_matrix = tf.convert_to_tensor(h_matrix.transpose((1, 0, 2)))
        else:
            self.h_matrix = tf.convert_to_tensor(
                h_matrix.transpose((1, 0, 2))[..., 1:-1]
            )

    def interp(self, m):
        p = self.point_value()
        p_r = tf.math.real(p)
        p_i = tf.math.imag(p)
        idx = self.get_bin_index(m)
        idx = tf.clip_by_value(idx, 0, self.h_matrix.shape[1] - 1)
        ret_r = do_spline_hmatrix(self.h_matrix, p_r, m, idx)
        ret_i = do_spline_hmatrix(self.h_matrix, p_i, m, idx)
        return tf.complex(ret_r, ret_i)


def do_spline_hmatrix(h_matrix, y, m, idx):
    ai, bi, ci, di = tf.unstack(tf.reduce_sum(h_matrix * y, axis=-1), axis=0)
    a, b, c, d = (
        tf.gather(ai, idx),
        tf.gather(bi, idx),
        tf.gather(ci, idx),
        tf.gather(di, idx),
    )
    ret = a + m * (b + m * (c + d * m))
    return ret


@register_particle("interp_l3")
class InterpL3(InterpolationParticle):
    def interp(self, m):
        p = self.point_value()
        ones = tf.ones_like(m)
        zeros = tf.zeros_like(m)
        p_r = tf.math.real(p)
        p_i = tf.math.imag(p)
        h, b = get_matrix_interp1d3_v2(m, self.points)
        h = tf.stop_gradient(h)
        f = lambda x: tf.reshape(
            tf.matmul(tf.cast(h, x.dtype), tf.reshape(x, (-1, 1))), b.shape
        ) + tf.cast(b, x.dtype)
        ret_r = f(p_r)
        ret_i = f(p_i)
        return tf.complex(ret_r, ret_i)


def get_matrix_interp1d3_v2(x, xi):
    N = len(xi) - 1
    zeros = tf.zeros_like(x)
    ones = tf.ones_like(x)
    # @pysnooper.snoop()
    def poly_i(i):
        tmp = zeros
        x_i = (xi[i] + xi[i - 1]) / 2
        for j in range(i - 1, i + 3):
            if j < 0 or j > N - 1:
                continue
            r = ones
            for k in range(j - 1, j + 3):
                if k == i or k < 1 or k > N:
                    continue
                x_k = (xi[k] + xi[k - 1]) / 2
                r = r * (x - x_k) / (x_i - x_k)
            r = tf.where(
                (x >= (xi[j] + xi[j - 1]) / 2) & (x < (xi[j] + xi[j + 1]) / 2),
                r,
                zeros,
            )
            tmp = tmp + r
        return tmp

    h = tf.stack([poly_i(i) for i in range(1, N)], axis=-1)
    b = tf.zeros_like(x)
    return h, b



import numpy as np

from tf_pwa.amp import get_particle, interpolation, variable_scope


def test_hist_idx():
    p = get_particle("a", model="hist_idx", min_m=1.0, max_m=3.0, interp_N=20)
    with variable_scope() as vm:
        p.init_params()
        vm.set("a_point_0r", 1.0)
        vm.set("a_point_0i", 0.0)
        vm.set("a_point_1r", 1.0)
        vm.set("a_point_1i", 0.0)
    amp = p(np.array([1.0, 1.001, 2.0, 3.0, 4.0]))
    assert np.allclose(amp[0:2], 1.0)


def test_spline_c_idx():
    p = get_particle(
        "a",
        model="spline_c_idx",
        min_m=1.0,
        max_m=3.0,
        interp_N=20,
        with_bound=True,
    )
    with variable_scope() as vm:
        p.init_params()
        for i in range(20):
            vm.set(f"a_point_{i}r", 1.0)
            vm.set(f"a_point_{i}i", 0.0)
        amp = p(np.array([1.0, 3.0, 2.0]))
    assert np.allclose(amp, 1.0)

    p = get_particle(
        "a",
        model="spline_c_idx",
        polar=False,
        points=[1.0, 2.0, 3.0, 4.0, 5.0],
    )
    with variable_scope() as vm:
        p.init_params()
        for i in range(3):
            vm.set(f"a_point_{i}r", i)
            vm.set(f"a_point_{i}i", i)
        amp = p(np.array([1.0, 3.0, 2.0, 4.0]))

    assert np.allclose(amp, [0.0, 1 + 1j, 0.0, 2 + 2j])



import pytest

from tf_pwa.amp import *
from tf_pwa.cal_angle import cal_angle_from_momentum
from tf_pwa.phasespace import PhaseSpaceGenerator


def generate_mc(num):
    a = PhaseSpaceGenerator(4.59925, [2.01026, 0.13957061, 2.00685])
    flat_mc_data = a.generate(num)
    return flat_mc_data


@pytest.mark.benchmark(group="generate_mc")
def test_generate_mc_CPU(benchmark):
    with tf.device("CPU:0"):
        benchmark(generate_mc, 1000000)


@pytest.mark.benchmark(group="generate_mc")
def test_generate_mc(benchmark):
    benchmark(generate_mc, 1000000)


def _get_decay_data(num=1):
    data = generate_mc(200000)
    a = Particle("A", J=1, P=-1, spins=(-1, 1))
    b = Particle("B", J=1, P=-1)
    c = Particle("C", J=0, P=-1)
    d = Particle("D", J=1, P=-1)
    res = Particle("res", 1, 1, mass=4.4232, width=0.025)
    dec1 = HelicityDecay(a, [res, c])
    dec2 = HelicityDecay(res, [b, d])
    if num == 1:
        decs = DecayGroup([[dec1, dec2]])
    else:
        res2 = Particle("res2", 1, 1, mass=2.4232, width=0.025)
        dec3 = HelicityDecay(a, [res2, d])
        dec4 = HelicityDecay(res2, [b, c])
        decs = DecayGroup([[dec1, dec2], [dec3, dec4]])
    p = dict(zip([d, b, c], data))
    amp = AmplitudeModel(decs)
    data = cal_angle_from_momentum(p, decs)
    args = amp.trainable_variables
    return amp, data, args


@pytest.mark.benchmark(group="mc_int")
def test_mc_int(benchmark):
    amp, data, params = _get_decay_data(1)

    def mc_int(dat):
        return tf.reduce_sum(amp(dat))

    benchmark(mc_int, data)


@pytest.mark.benchmark(group="mc_int")
def test_mc_int_CPU(benchmark):
    amp, data, params = _get_decay_data(1)

    def mc_int(dat):
        return tf.reduce_sum(amp(dat))

    with tf.device("CPU:0"):
        benchmark(mc_int, data)


@pytest.mark.benchmark(group="mc_int")
def test_mc_int2(benchmark):
    amp, data, params = _get_decay_data(2)

    def mc_int(dat):
        return tf.reduce_sum(amp(dat))

    benchmark(mc_int, data)


@pytest.mark.benchmark(group="mc_int")
def test_mc_int2_CPU(benchmark):
    amp, data, params = _get_decay_data(2)

    def mc_int(dat):
        return tf.reduce_sum(amp(dat))

    with tf.device("CPU:0"):
        benchmark(mc_int, data)


@pytest.mark.benchmark(group="mc_int")
def test_mc_int_grad(benchmark):
    amp, data, params = _get_decay_data(1)

    def mc_int(dat):
        with tf.GradientTape() as tape:
            int_mc = tf.reduce_sum(amp(dat))
        args = tape.gradient(int_mc, params)
        return int_mc, args

    benchmark(mc_int, data)


@pytest.mark.benchmark(group="mc_int")
def test_mc_int_grad_CPU(benchmark):
    amp, data, params = _get_decay_data(1)

    def mc_int(dat):
        with tf.GradientTape() as tape:
            int_mc = tf.reduce_sum(amp(dat))
        args = tape.gradient(int_mc, params)
        return int_mc, args

    with tf.device("CPU:0"):
        benchmark(mc_int, data)



#!/usr/bin/env python3
import copy
import os.path
import sys
from pprint import pprint

import yaml

from tf_pwa.config_loader import ConfigLoader, MultiConfig
from tf_pwa.significance import significance


def single_fit(configs, res, data):
    config = MultiConfig(configs, share_dict=res)

    print("\n########### initial parameters")
    pprint(config.get_params())
    for i in config.configs:
        print("decay chain:")
        for j in i.full_decay:
            print(j)
    fit_result = config.fit(data)
    pprint(fit_result.params)
    # fit_result.save_as("final_params.json")
    return fit_result.min_nll, fit_result.ndf, fit_result.success


def multi_fit(configs, res, data, num=5):
    nll, ndf, status = single_fit(configs, res, data)
    for i in range(num - 1):
        nll_i, ndf_i, status_i = single_fit(configs, res, data)
        assert ndf_i == ndf
        if (not status) and status_i:
            nll = nll_i
            status = status_i
        else:
            if nll > nll_i:
                nll = nll_i
    return nll, ndf


def cached_data(config_dict):
    config = ConfigLoader(config_dict)
    data = config.get_all_data()
    return data


def load_config(f_name):
    with open(f_name) as f:
        ret = yaml.safe_load(f)
    return ret


def cal_significance(config_files, res_file_name, base_res, test_res, loop=5):
    configs = [load_config(i) for i in config_files]
    res_file = load_config(res_file_name)

    for i in res_file:
        res_file[i]["disable"] = True

    data = [cached_data(i) for i in configs]

    def get_config(r):
        tmp_res = copy.deepcopy(res_file)
        for i in base_res:
            tmp_res[i]["disable"] = False
        for i in r:
            if i in base_res:
                tmp_res[i]["disable"] = True
            else:
                tmp_res[i]["disable"] = False
        return {res_file_name: tmp_res}

    base_config = get_config([])
    nll, ndf = multi_fit(configs, base_config, data, num=loop)
    nlls = {"base": nll}
    ndfs = {"base": ndf}
    print("nll: {}, ndf: {}".format(nll, ndf))
    signi = {}
    for i in test_res:
        print("\ncalculate significance for {}\n".format(i))
        config_i = get_config([i])

        nll_i, ndf_i = multi_fit(configs, config_i, data)
        nlls[i] = nll_i
        ndfs[i] = ndf_i
        signi[i] = significance(nll, nll_i, abs(ndf - ndf_i))
        print(
            "nll: {}, ndf: {}, significane: {}".format(nll_i, ndf_i, signi[i])
        )
    return signi, nlls, ndfs


def main():
    res = [
        "X(3940)(1+)",
        "X(3940)(1-)",
        "X(3940)(0-)",
        "X(3940)(2+)",
        "X(3940)(2-)",
    ]

    config_files = [
        "config.yml",
        "config_1.yml",
        "config_K3pi_1.yml",
        "config_K3pi_2.yml",
    ]
    res_files = "Resonances_B.yml"

    base_res = [
        "Psi(4660)",
        "Psi(4230)",
        "Psi(4390)",
        "Psi(4260)",
        "Psi(4360)",
    ]

    test_res = base_res  # [""]

    signi, nlls, ndfs = cal_significance(
        config_files, res_files, base_res, test_res, loop=5
    )
    print("base", nlls["base"], ndfs["base"])
    print("particle\tsignificance\tnll\tndf")
    for i in signi:
        print(i, signi[i], nlls[i], ndfs[i])


if __name__ == "__main__":
    main()



import numpy as np
import pytest

from tf_pwa.data import *
from tf_pwa.particle import BaseDecay, BaseParticle

from .common import write_temp_file


def test_index():
    data = {"s": [{"s": 1}, {"d": 2}], "p": (3, 4)}
    assert data_index(data, ("s", 1)) == {"d": 2}
    assert data_index(data, ["p", 0]) == 3
    assert data_index(data, "p") == (3, 4)
    with pytest.raises(ValueError):
        data_index(data, "t")


def test_index2():
    a = BaseParticle("A")
    b = BaseParticle("B")
    c = BaseParticle("C")
    d = BaseDecay(a, [b, c])
    data = {
        "particle": {a: {"p": 1}, b: {"m": 2}},
        "decay": [
            {d: {b: {"a": 3}, c: {"a": 4}}},
        ],
    }
    assert data_index(data, ["particle", a, "p"]) == 1
    assert data_index(data, ["particle", "B", "m"]) == 2
    assert data_index(data, ["decay", 0, str(d), b, "a"]) == 3
    assert data_index(data, ["decay", 0, d, str(c), BaseParticle("a")]) == 4


def test_flatten_dict_data():
    a = BaseParticle("A")
    b = BaseParticle("B")
    c = BaseParticle("C")
    d = BaseDecay(a, [b, c])
    data = {
        "particle": {a: {"p": 1}, b: {"m": 2}},
        "decay": [
            {d: {b: {"a": 3}, c: {"a": 4}}},
        ],
    }
    data_f = flatten_dict_data(data)
    assert len(data_f) == 4
    assert data_f["particle/A/p"] == 1
    assert data_f["particle/B/m"] == 2
    assert data_f["decay/0/{}/B/a".format(d)] == 3
    assert data_f["decay/0/{}/C/a".format(d)] == 4


def test_load_dat_file():
    s1 = "5.0 2.0 3.0 3.0\n"
    s2 = "4.0 2.0 4.0 2.0\n4.0 2.0 4.0 2.0\n"
    print(s1 + s2)
    with write_temp_file(s1 + s2) as fname:
        dat1 = load_dat_file(fname, ["a", "b", "c"])
    with write_temp_file(s1) as fname1:
        with write_temp_file(s2) as fname2:
            dat2 = load_dat_file([fname1, fname2], ["a", "b", "c"])
    assert np.allclose(dat1["a"], dat2["a"])
    assert np.allclose(dat1["b"], dat2["b"])
    assert np.allclose(dat1["c"], dat2["c"])


def test_save_load():
    b = BaseParticle("b")
    data = {"a": np.array([1.0, 2.0, 3.0]), b: np.array([2.0, 3.0, 4.0])}

    fname = "test_save_load.npy"
    save_data(fname, data)
    data2 = load_data(fname)
    assert np.allclose(data["a"], data2["a"])
    assert np.allclose(data[b], data2[b])


def test_set_seed():
    set_random_seed(100)
    a = np.random.random((10,))
    set_random_seed(100)
    b = np.random.random((10,))
    assert np.allclose(a, b)



import json
import os.path
import pickle
import sys
import time

import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf
from fit_scipy_new import (
    cal_hesse_error,
    get_amplitude,
    get_decay_chains,
    load_params,
)
from scipy.optimize import minimize

from tf_pwa.applications import (
    compare_result,
    fit_fractions,
    gen_data,
    gen_mc,
    plot_pull,
)
from tf_pwa.cal_angle import prepare_data_from_decay
from tf_pwa.data import data_to_tensor, load_dat_file, split_generator
from tf_pwa.fitfractions import cal_fitfractions, cal_fitfractions_no_grad

# for fit
from tf_pwa.model import FCN, Model
from tf_pwa.utils import error_print, load_config_file, pprint

this_dir = os.path.dirname(__file__)
sys.path.insert(0, this_dir + "/..")


def prepare_data(fname, decs, particles=None, dtype="float64"):
    data_np = prepare_data_from_decay(
        fname, decs, particles=particles, dtype=dtype
    )
    return data_to_tensor(data_np)


def gen_data_from_mc():
    config_list = load_config_file("Resonances")
    decs, final_particles, decay = get_decay_chains(config_list)
    amp = get_amplitude(decs, config_list, decay)
    load_params(amp, "gen_params.json")
    gen_mc(4.59925172, [2.00698, 2.01028, 0.13957], 10000, "data/flat_mc.dat")
    gen_data(
        amp,
        final_particles,
        Ndata=10000,
        mcfile="data/flat_mc.dat",
        genfile="data/gen_data.dat",
        Poisson_fluc=True,
    )


def fit():
    dtype = "float64"

    config_list = load_config_file("Resonances")
    decs, final_particles, decay = get_decay_chains(config_list)
    amp = get_amplitude(decs, config_list, decay)
    # data/pure1w.dat  data/flat_mc30w.dat
    data = prepare_data(
        "data/data4600_new.dat", decs, particles=final_particles, dtype=dtype
    )
    mcdata = prepare_data(
        "data/PHSP4600_new.dat", decs, particles=final_particles, dtype=dtype
    )
    bg = prepare_data(
        "data/bg4600_new.dat", decs, particles=final_particles, dtype=dtype
    )
    w_bkg = 0.768331

    model = Model(amp, w_bkg=w_bkg)
    load_params(model.Amp, "glb_params_rp.json")
    model.Amp.vm.trans_params(True)
    args_name = model.Amp.vm.trainable_vars

    fcn = FCN(model, data, mcdata, bg, batch=65000)
    SCIPY = False
    hesse = True
    if SCIPY:

        def callback(x):
            print(fcn.cached_nll)

        f_g = fcn.model.Amp.vm.trans_fcn_grad(fcn.nll_grad)
        now = time.time()
        s = minimize(
            f_g,
            np.array(fcn.model.Amp.vm.get_all_val(True)),
            method="BFGS",
            jac=True,
            callback=callback,
            options={"disp": 1, "gtol": 1e-4, "maxiter": 1000},
        )
        model.Amp.vm.trans_params(True)
        xn = fcn.model.Amp.vm.get_all_val()
        print("########## fit state:")
        print(s)
        print("\nTime for fitting:", time.time() - now)
        val = {k: v.numpy() for k, v in model.Amp.variables.items()}
        err = {}
        if hesse:
            inv_he = cal_hesse_error(
                model.Amp, val, w_bkg, data, mcdata, bg, args_name, batch=20000
            )
            diag_he = inv_he.diagonal()
            hesse_error = np.sqrt(np.fabs(diag_he)).tolist()
            print(hesse_error)
            err = dict(zip(model.Amp.vm.trainable_vars, hesse_error))

    else:
        from tf_pwa.fit import fit_minuit

        m = fit_minuit(fcn, hesse=False)
        model.Amp.vm.trans_params(True)
        err_mtrx = m.np_covariance()
        np.save("error_matrix.npy", err_mtrx)
        val = dict(m.values)
        err = dict(m.errors)

    print("\n########## fit results:")
    for i in val:
        print("  ", i, ":", error_print(val[i], err.get(i, None)))

    frac = True
    err_frac = {}
    if frac:
        err_frac = {}
        if hesse:
            frac, grad = cal_fitfractions(
                model.Amp, list(split_generator(mcdata, 25000))
            )
        else:
            frac = cal_fitfractions_no_grad(
                model.Amp, list(split_generator(mcdata, 45000))
            )
        for i in frac:
            if hesse:
                err_frac[i] = np.sqrt(np.dot(np.dot(inv_he, grad[i]), grad[i]))
        print("########## fit fractions")
        for i in frac:
            print(i, ":", error_print(frac[i], err_frac.get(i, None)))

    outdic = {
        "value": val,
        "error": err,
        "config": config_list,
        "frac": frac,
        "err_frac": err_frac,
        "NLL": fcn.cached_nll.numpy(),
    }
    with open("glb_params_rp.json", "w") as f:
        json.dump(outdic, f, indent=2)


def gen_toy_sample():
    Ndata = 10000  # 8065
    Nbg = 0  # 3445
    w_bkg = 0  # .768331
    mcfile = "data/flat_mc30w.dat"
    bgfile = "data/bg4600_new.dat"
    params_file = "gen_params_rp.json"

    dtype = "float64"
    config_list = load_config_file("Resonances")
    decs, final_particles, decay = get_decay_chains(config_list)
    amp = get_amplitude(decs, config_list, decay)

    if w_bkg == 0:
        bg = None
        bgfile = None
    else:
        bg = prepare_data(bgfile, decs, particles=final_particles, dtype=dtype)
    mcdata = prepare_data(mcfile, decs, particles=final_particles, dtype=dtype)
    model = Model(amp, w_bkg=w_bkg)
    args_name = model.Amp.vm.trainable_vars

    var_arr = []
    err_arr = []
    frac_arr = []
    err_frac_arr = []
    for number in range(200):
        model.Amp.vm.trans_params(True)  # switch to rp

        load_params(model.Amp, params_file)
        # model.Amp.vm.trans_params(True)
        # gen_mc(4.59925172,[2.00698,2.01028,0.13957],60000,"data/flat_mc.dat")
        data = gen_data(
            model.Amp,
            final_particles,
            Ndata,
            mcfile=mcfile,
            Poisson_fluc=True,
            Nbg=Nbg,
            wbg=w_bkg,
            bgfile=bgfile,
        )

        fcn = FCN(model, data, mcdata, bg=bg, batch=65000)

        def callback(x):
            print(fcn.cached_nll)

        # fcn.model.Amp.vm.set_bound(bounds_dict)
        f_g = fcn.model.Amp.vm.trans_fcn_grad(fcn.nll_grad)
        now = time.time()
        s = minimize(
            f_g,
            np.array(fcn.model.Amp.vm.get_all_val(True)),
            method="BFGS",
            jac=True,
            callback=callback,
            options={"disp": 1, "gtol": 1e-4, "maxiter": 1000},
        )
        fcn.model.Amp.vm.set_all(s.x)
        model.Amp.vm.trans_params(False)  # switch to xy
        var = fcn.model.Amp.vm.get_all_dic(trainable_only=True)
        var_all = {k: v.numpy() for k, v in model.Amp.variables.items()}
        print("\nTime for fitting:", time.time() - now)
        print("fitting parameters:\n", var)

        err = {}
        inv_he = cal_hesse_error(
            model.Amp, var_all, w_bkg, data, mcdata, bg, args_name, batch=20000
        )
        diag_he = inv_he.diagonal()
        hesse_error = np.sqrt(np.fabs(diag_he)).tolist()
        err = dict(zip(model.Amp.vm.trainable_vars, hesse_error))
        print("errors:\n", err)

        frac, err_frac = fit_fractions(model, mcdata, inv_he, hesse=True)
        print("fitfractions:\n", frac)
        print("errors:\n", err_frac)

        var_arr.append(var)
        err_arr.append(err)
        frac_arr.append(frac)
        err_frac_arr.append(err_frac)
        if not s.success:
            print("NOT SUCCESSFUL FITTING")
        print("\n{0}END in {1}s\n".format(number, time.time() - now))

    print("$$$$$ var\n", var_arr)
    print("$$$$$ err\n", err_arr)
    print("$$$$$ frac\n", frac_arr)
    print("$$$$$ err_frac\n", err_frac_arr)

    output = open("toyMar9.pkl", "wb")
    pickle.dump(
        {
            "var": var_arr,
            "err": err_arr,
            "frac": frac_arr,
            "err_frac": err_frac_arr,
        },
        output,
        -1,
    )
    output.close()


def compare_toy():
    with open("toyMar9.pkl", "rb") as f:
        dd = pickle.load(f)
    var = dd["var"]  # [{var_name:var_val}]
    err = dd["err"]
    frac = dd["frac"]  # [{frac_name:frac_val}]
    err_frac = dd["err_frac"]
    with open("gen_params_xy.json") as f:
        result = json.load(f)

    POLAR = False
    if POLAR:
        periodic_vars = []
        for v in result["error"]:
            if v[-1] == "i":
                periodic_vars.append(v)
    else:
        periodic_vars = [
            "A->D2_2460p.BD2_2460p->D.C_totali",
            "A->D1_2420.DD1_2420->B.C_totali",
            "A->D1_2420p.BD1_2420p->D.C_totali",
            "A->D1_2430.DD1_2430->B.C_totali",
            "A->D1_2430p.BD1_2430p->D.C_totali",
        ]

    for i in range(var.__len__()):
        print("##### Pull", i)
        compare_result(
            var[i],
            result["value"],
            err[i],  # result["error"],
            figname="fig/tmp/var_pull_{}".format(i),
            yrange=8,
            periodic_vars=periodic_vars,
        )
        """compare_result(result["frac"], frac[i], result["err_frac"], err_frac[i],
                       figname="fig/tmp/frac_pull_{}".format(i), yrange=5)"""


def toy_pull():
    with open("toyMar10.pkl", "rb") as f:
        dd = pickle.load(f)
    var = dd["var"]  # [{var_name:var_val}]
    err = dd["err"]
    frac = dd["frac"]  # [{frac_name:frac_val}]
    err_frac = dd["err_frac"]
    with open("gen_params_xy.json") as f:
        result = json.load(f)
    del_list = []
    var = np.delete(var, del_list, axis=0)
    err = np.delete(err, del_list, axis=0)
    frac = np.delete(frac, del_list, axis=0)
    err_frac = np.delete(err_frac, del_list, axis=0)

    vv = {}
    err_vv = {}
    for i in var[0]:
        vv[i] = []
        err_vv[i] = []
    for i in vv:
        for j in var:
            vv[i].append(j[i])
        for j in err:
            err_vv[i].append(j[i])
    var = vv  # {var_name:[var_val]}
    err = err_vv

    ff = {}
    err_ff = {}
    for i in frac[0]:
        ff[i] = []
        err_ff[i] = []
    for i in ff:
        for j in frac:
            ff[i].append(j[i])
        for j in err_frac:
            err_ff[i].append(j[i])
    frac = ff  # {frac_name:[frac_val]}
    err_frac = err_ff

    m = []
    s = []
    e_m = []
    e_s = []
    for i in var:
        mean, sigma, err_m, err_s = plot_pull(
            var[i],
            "var" + i,
            norm=True,
            value=result["value"][i],
            error=err[i],
        )
        m.append(mean)
        s.append(sigma)
        e_m.append(err_m)
        e_s.append(err_s)
        print("Plot pull of {}".format(i))
    print("mean", m)
    print("err_mean", e_m)
    print("sigma", s)
    print("err_sigma", e_s)

    m = []
    s = []
    e_m = []
    e_s = []
    for i in frac:
        mean, sigma, err_m, err_s = plot_pull(
            frac[i],
            "frac" + i,
            norm=True,
            value=result["frac"][i],
            error=err_frac[i],
        )
        m.append(mean)
        s.append(sigma)
        e_m.append(err_m)
        e_s.append(err_s)
        print("Plot pull of {}".format(i))
    print("mean", m)
    print("err_mean", e_m)
    print("sigma", s)
    print("err_sigma", e_s)


def draw_errorbar():
    mu = []
    err_mu = []
    plt.errorbar(
        range(len(mu)),
        mu,
        yerr=err_mu,
        fmt="oy",
        ecolor="r",
        elinewidth=1,
        capsize=3,
    )
    plt.errorbar(range(len(mu)), 0, fmt="g")
    plt.xlabel("var index")
    plt.ylabel("value")
    plt.title("mu")
    plt.show()
    # plt.savefig("errbar_mu")
    # plt.clf()

    sigma = []
    err_sigma = []
    plt.errorbar(
        range(len(sigma)),
        sigma,
        yerr=err_sigma,
        fmt="oy",
        ecolor="r",
        elinewidth=1,
        capsize=3,
    )
    plt.errorbar(range(len(sigma)), 1, fmt="g")
    plt.xlabel("var index")
    plt.ylabel("value")
    plt.title("sigma")
    plt.show()
    # plt.savefig("errbar_sigma")
    # plt.clf()


if __name__ == "__main__":
    toy_pull()



import numpy as np

from tf_pwa import root_io


def test_read_load():
    dic = {"a": [1.0, 2.03]}
    root_io.save_dict_to_root(dic, "test_io.root", "data")
    dic2 = root_io.load_root_data("test_io.root")
    assert np.allclose(dic2["data0"]["a"].numpy(), [1.0, 2.03])



from tf_pwa.tensorflow_wrapper import tf

from .core import Particle, Variable, register_particle


def cal_monentum(m, ma, mb):
    mabp = ma + mb
    mabm = ma - mb
    s = m * m
    p2 = (s - mabp * mabp) * (s - mabm * mabm) / 4 / s
    zeros = tf.zeros_like(s)
    p_p = tf.complex(tf.sqrt(tf.abs(p2)), zeros)
    p_m = tf.complex(zeros, tf.sqrt(tf.abs(p2)))
    return tf.where(p2 > 0, p_p, p_m)


@register_particle("Flatte")
class ParticleFlatte(Particle):
    def __init__(self, *args, mass_list=None, **kwargs):
        super().__init__(*args, **kwargs)
        if mass_list is None:
            raise ValueError("required mass_list: [[a, b], [mc, md]]")
        self.mass_list = mass_list
        self.g_value = []
        self.float_list = list(kwargs.get("float", []))

    def init_params(self):
        self.d = 3.0
        if self.mass is None:
            self.mass = self.add_var("mass", fix=True)
            # print("$$$$$",self.mass)
        else:
            if not isinstance(self.mass, Variable):
                if "m" in self.float_list:
                    self.mass = self.add_var(
                        "mass", value=self.mass, fix=False
                    )
                else:
                    self.mass = self.add_var("mass", value=self.mass, fix=True)
        self.g_value = []
        for i, mab in enumerate(self.mass_list):
            name = f"g_{i}"
            if hasattr(self, name):
                if name in self.float_list:
                    self.g_value.append(
                        self.add_var(
                            f"g_{i}", value=getattr(self, name), fix=False
                        )
                    )
                else:
                    self.g_value.append(
                        self.add_var(
                            f"g_{i}", value=getattr(self, name), fix=True
                        )
                    )
            else:
                self.g_value.append(self.add_var(f"g_{i}"))

    def __call__(self, m):
        return self.get_amp({"m": m})

    def get_amp(self, *args, **kwargs):
        m = args[0]["m"]
        mass = self.get_mass()
        zeros = tf.zeros_like(m)
        delta_s = mass * mass - m * m
        m_c = mass / m
        rhos = []
        for i, mab in enumerate(self.mass_list):
            ma, mb = mab
            pi = cal_monentum(m, ma, mb)
            # print(pi)
            m_rho_i = pi * tf.complex(zeros, self.g_value[i]() * m_c)
            rhos.append(m_rho_i)
        rho = sum(rhos)
        re = delta_s + tf.math.real(rho)
        im = tf.math.imag(rho)
        d = re * re + im * im
        ret = tf.complex(re / d, -im / d)
        return ret



#!/usr/bin/env python3
import datetime
import json
import os
import os.path
import sys
from functools import reduce
from math import pi

import matplotlib.pyplot as plt
from fit_scipy_new import (
    get_amplitude,
    get_decay_chains,
    load_params,
    prepare_data,
)
from scipy import interpolate

import tf_pwa
from tf_pwa.amp import *
from tf_pwa.cal_angle import prepare_data_from_decay
from tf_pwa.data import flatten_dict_data
from tf_pwa.utils import load_config_file

this_dir = os.path.dirname(__file__)
sys.path.insert(0, this_dir + "/..")


def config_split(config):
    ret = {}
    for i in config:
        ret[i] = {i: config[i]}
    return ret


def part_config(config, name=None):
    name = name if name is not None else []
    if isinstance(name, str):
        print(name)
        return {name: config[name]}
    ret = {}
    for i in name:
        if i in config:
            ret[i] = config[i]
    return ret


def equal_pm(a, b):
    def remove_pm(s):
        ret = s
        if s.endswith("p"):
            ret = s[:-1]
        elif s.endswith("m"):
            ret = s[:-1]
        return ret

    return remove_pm(a) == remove_pm(b)


def part_combine_pm(config_list):
    ret = []
    for i in config_list:
        for j in ret:
            if equal_pm(i, j[0]):
                j.append(i)
                break
        else:
            ret.append([i])
    return ret


def hist_line(data, weights, bins, xrange=None, inter=1):
    y, x = np.histogram(data, bins=bins, range=xrange, weights=weights)
    x = (x[:-1] + x[1:]) / 2
    func = interpolate.interp1d(x, y, kind="quadratic")
    delta = (xrange[1] - xrange[0]) / bins / inter
    x_new = np.arange(x[0], x[-1], delta)
    y_new = func(x_new)
    return x_new, y_new


param_list_test = [
    "particle/A/m",
    "particle/B/m",
    "particle/C/m",
    "particle/D/m",
    "particle/(B, C)/m",
    "particle/(B, D)/m",
    "particle/(C, D)/m",
    "decay/A->(B, C)+D/(B, C)/ang/beta",
    "decay/(B, C)->B+C/B/ang/beta",
    "decay/A->(B, C)+D/(B, C)/ang/alpha",
    "decay/(B, C)->B+C/B/ang/alpha",
    "decay/A->(B, D)+C/(B, D)/ang/beta",
    "decay/(B, D)->B+D/B/ang/beta",
    "decay/A->(B, D)+C/(B, D)/ang/alpha",
    "decay/(B, D)->B+D/B/ang/alpha",
    "decay/A->(C, D)+B/(C, D)/ang/beta",
    "decay/(C, D)->D+C/D/ang/beta",
    "decay/A->(C, D)+B/(C, D)/ang/alpha",
    "decay/(C, D)->D+C/D/ang/alpha",
    "decay/(B, D)->B+D/B/aligned_angle/beta",
    "decay/(B, C)->B+C/B/aligned_angle/beta",
    "decay/(B, D)->B+D/D/aligned_angle/beta",
    "decay/(C, D)->D+C/D/aligned_angle/beta",
    "decay/(B, D)->B+D/B/aligned_angle/alpha",
    "decay/(B, D)->B+D/B/aligned_angle/gamma",
    "decay/(B, C)->B+C/B/aligned_angle/alpha",
    "decay/(B, C)->B+C/B/aligned_angle/gamma",
    "decay/(B, D)->B+D/D/aligned_angle/alpha",
    "decay/(B, D)->B+D/D/aligned_angle/gamma",
    "decay/(C, D)->D+C/D/aligned_angle/alpha",
    "decay/(C, D)->D+C/D/aligned_angle/gamma",
]

param_list = [
    "m_A",
    "m_B",
    "m_C",
    "m_D",
    "m_BC",
    "m_BD",
    "m_CD",
    "beta_BC",
    "beta_B_BC",
    "alpha_BC",
    "alpha_B_BC",
    "beta_BD",
    "beta_B_BD",
    "alpha_BD",
    "alpha_B_BD",
    "beta_CD",
    "beta_D_CD",
    "alpha_CD",
    "alpha_D_CD",
    "beta_BD_B",
    "beta_BC_B",
    "beta_BD_D",
    "beta_CD_D",
    "alpha_BD_B",
    "gamma_BD_B",
    "alpha_BC_B",
    "gamma_BC_B",
    "alpha_BD_D",
    "gamma_BD_D",
    "alpha_CD_D",
    "gamma_CD_D",
]

params_config = {
    "m_BC": {
        "xrange": (2.15, 2.65),
        "display": "$m_{ {D*}^{-}\pi^{+} }$",
        "bins": 50,
        "units": "Gev",
    },
    "m_BD": {
        "xrange": (4.0, 4.47),
        "display": "$m_{ {D*}^{-}{D*}^{0} }$",
        "bins": 47,
        "units": "GeV",
    },
    "m_CD": {
        "xrange": (2.15, 2.65),
        "display": "$m_{ {D*}^{0}\pi^{+} }$",
        "bins": 50,
        "units": "GeV",
    },
    "alpha_B_BD": {
        "xrange": (-pi, pi),
        "display": r"$\phi^{ {D*}^{-} }_{ {D*}^{0} {D*}^{-} }$",
        "bins": 50,
        "units": "",
        "legend": False,
    },
    "alpha_BD": {
        "xrange": (-pi, pi),
        "display": r"$ \phi_{ {D*}^{0} {D*}^{-} }$",
        "bins": 50,
        "units": "",
        "legend": False,
    },
    "cosbeta_BD": {
        "xrange": (-1, 1),
        "display": r"$\cos \theta_{ {D*}^{0} {D*}^{-} }$",
        "bins": 50,
        "units": "",
        "legend": False,
    },
    "cosbeta_B_BD": {
        "xrange": (-1, 1),
        "display": r"$\cos \theta^{ {D*}^{-} }_{ {D*}^{0} {D*}^{-} }$",
        "bins": 50,
        "units": "",
        "legend": False,
    },
    "alpha_B_BC": {
        "xrange": (-pi, pi),
        "display": r"$\phi^{ {D*}^{-} }_{ {D*}^{-}\pi^{+} }$",
        "bins": 50,
        "units": "",
        "legend": False,
    },
    "alpha_BC": {
        "xrange": (-pi, pi),
        "display": r"$ \phi_{ {D*}^{-} \pi^{+} }$",
        "bins": 50,
        "units": "",
        "legend": False,
    },
    "cosbeta_BC": {
        "xrange": (-1, 1),
        "display": r"$\cos \theta_{ {D*}^{-} \pi^{+} }$",
        "bins": 50,
        "units": "",
        "legend": False,
    },
    "cosbeta_B_BC": {
        "xrange": (-1, 1),
        "display": r"$\cos \theta^{ {D*}^{-} }_{ {D*}^{-}\pi^{+} }$",
        "bins": 50,
        "units": "",
        "legend": False,
    },
    "alpha_D_CD": {
        "xrange": (-pi, pi),
        "display": r"$\phi^{ {D*}^{0} }_{ {D*}^{0}\pi^{+} }$",
        "bins": 50,
        "units": "",
        "legend": False,
    },
    "alpha_CD": {
        "xrange": (-pi, pi),
        "display": r"$ \phi_{ {D*}^{0} \pi^{+} }$",
        "bins": 50,
        "units": "",
        "legend": False,
    },
    "cosbeta_CD": {
        "xrange": (-1, 1),
        "display": r"$\cos \theta_{ {D*}^{0} \pi^{+} }$",
        "bins": 50,
        "units": "",
        "legend": False,
    },
    "cosbeta_D_CD": {
        "xrange": (-1, 1),
        "display": r"$\cos \theta^{ {D*}^{0} }_{ {D*}^{0}\pi^{+} }$",
        "bins": 50,
        "units": "",
        "legend": False,
    },
}
params_list_map = dict(zip(param_list, param_list_test))


def map_idx_for_params():
    for i in range(len(param_list)):
        name = o_name = param_list[i]
        if name.startswith("beta"):
            name = "cos" + name
        if name in params_config:
            params_config[name]["idx"] = params_list_map[o_name]


map_idx_for_params()


def get_weight(
    amp, config_list, mcdata, norm_int=1.0, res_list=None, pm_combine=None
):
    a_weight = {}
    if res_list is None:
        if pm_combine:
            res_list = part_combine_pm(config_list)
        else:
            res_list = [[i] for i in config_list]
    config_res = [part_config(config_list, i) for i in res_list]
    res_name = {}
    for i in range(len(res_list)):
        name = res_list[i]
        if isinstance(name, list):
            if len(name) > 1:
                name = reduce(lambda x, y: "{}+{}".format(x, y), res_list[i])
            else:
                name = name[0]
        res_name[i] = name
        amp.set_used_res(config_res[i])
        a_weight[i] = amp(mcdata, cached=True).numpy() * norm_int
    return a_weight, res_list, res_name


def get_idx_data(data, idx):
    data_idx = flatten_dict_data(data)
    return data_idx[idx].numpy()


def plot(params="final_params.json", res_list=None, pm_combine=True):
    dtype = "float64"
    w_bkg = 0.768331

    config_list = load_config_file("Resonances")

    decs, final_particles, decay = get_decay_chains(config_list)
    data, bg, mcdata = prepare_data(
        decs, particles=final_particles, dtype=dtype
    )
    amp = get_amplitude(decs, config_list, decay)
    load_params(amp, params)

    total = amp(mcdata)
    int_mc = tf.reduce_sum(total).numpy()
    n_data = data_shape(data)
    n_bg = data_shape(bg)
    norm_int = (n_data - w_bkg * n_bg) / int_mc

    a_weight, res_list, res_name = get_weight(
        amp, config_list, mcdata, norm_int, res_list, pm_combine
    )

    cmap = plt.get_cmap("jet")
    N = 3 + len(res_list)
    colors = [cmap(float(i) / N) for i in range(N)]
    colors = [
        "black",
        "red",
        "green",
        "blue",
        "yellow",
        "magenta",
        "cyan",
        "purple",
        "teal",
        "springgreen",
        "azure",
    ] + colors

    def plot_params(
        ax,
        name,
        bins=None,
        xrange=None,
        idx=0,
        display=None,
        units="GeV",
        legend=True,
    ):
        fd = lambda x: x
        if name.startswith("cos"):
            fd = lambda x: np.cos(x)
        inter = 2
        color = iter(colors)
        if display is None:
            display = name
        data_hist = np.histogram(
            fd(get_idx_data(data, idx)), range=xrange, bins=bins
        )
        # ax.hist(fd(data[idx].numpy()),range=xrange,bins=bins,histtype="step",label="data",zorder=99,color="black")
        data_y, data_x = data_hist[0:2]
        data_x = (data_x[:-1] + data_x[1:]) / 2
        data_err = np.sqrt(data_y)
        ax.errorbar(
            data_x,
            data_y,
            yerr=data_err,
            fmt=".",
            color=next(color),
            zorder=-2,
        )
        if bg is not None:
            ax.hist(
                fd(get_idx_data(bg, idx)),
                range=xrange,
                bins=bins,
                histtype="stepfilled",
                alpha=0.5,
                color="grey",
                weights=[w_bkg] * n_bg,
                label="bg",
                zorder=-1,
            )
            mc_bg = fd(
                np.append(get_idx_data(bg, idx), get_idx_data(mcdata, idx))
            )
            mc_bg_w = np.append([w_bkg] * n_bg, total.numpy() * norm_int)
        else:
            mc_bg = fd(get_idx_data(mcdata, idx))
            mc_bg_w = total.numpy() * norm_int
        x_mc, y_mc = hist_line(mc_bg, mc_bg_w, bins, xrange)
        # ax.plot(x_mc,y_mc,label="total fit")
        ax.hist(
            mc_bg,
            weights=mc_bg_w,
            bins=bins,
            range=xrange,
            histtype="step",
            color=next(color),
            label="total fit",
            zorder=100,
        )
        for i in a_weight:
            weights = a_weight[i]
            x, y = hist_line(
                fd(get_idx_data(mcdata, idx)), weights, bins, xrange, inter
            )
            y = y
            ax.plot(
                x,
                y,
                label=res_name[i],
                linestyle="solid",
                linewidth=1,
                color=next(color),
            )
        if legend:
            ax.legend(framealpha=0.5, fontsize="small")
        ax.set_ylabel("events/({:.3f} {})".format((x_mc[1] - x_mc[0]), units))
        ax.set_xlabel("{} {}".format(display, units))
        if xrange is not None:
            ax.set_xlim(xrange[0], xrange[1])
        ax.set_ylim(0, None)
        ax.set_title(display)

    plot_list = [
        "m_BC",
        "m_BD",
        "m_CD",
        "alpha_BD",
        "cosbeta_BD",
        "alpha_B_BD",
        "cosbeta_B_BD",
        "alpha_BC",
        "cosbeta_BC",
        "alpha_B_BC",
        "cosbeta_B_BC",
        "alpha_CD",
        "cosbeta_CD",
        "alpha_D_CD",
        "cosbeta_D_CD",
    ]
    n = len(plot_list)
    if not os.path.exists("figure"):
        os.mkdir("figure")
    # plt.style.use("classic")
    for i in range(n):
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)

        name = plot_list[i]
        plot_params(ax, name, **params_config.get(name, {}))
        fig.savefig("figure/" + name + ".pdf")
        fig.savefig("figure/" + name + ".png", dpi=300)


def plot_fig(ax, a):
    pass


if __name__ == "__main__":
    test_res_list = [
        ["Zc_4025"],
        ["Zc_4160"],
        ["D1_2420", "D1_2420p"],
        ["D1_2430", "D1_2430p"],
        ["D2_2460", "D2_2460p"],
    ]
    with tf.device("/device:CPU:0"):
        plot("final_params.json", res_list=None)



from math import pi

import numpy as np
import tensorflow as tf

from .angle import LorentzVector


def get_p(M, ma, mb):
    m2 = M * M
    m_p = (ma + mb) ** 2
    m_m = (ma - mb) ** 2
    p2 = (m2 - m_p) * (m2 - m_m)
    p = (p2 + tf.abs(p2)) / 2
    ret = tf.sqrt(p) / (2.0 * M)
    return tf.cast(ret, "float64")


class PhaseSpaceGenerator(object):
    """Phase Space Generator for n-body decay"""

    def __init__(self, m0, mass):
        self.m_mass = []
        self.set_decay(m0, mass)
        self.sum_mass = sum(self.m_mass)

    def generate_mass(self, n_iter):
        """generate possible inner mass."""
        sm = self.sum_mass - self.m_mass[-1] - self.m_mass[-2]
        m_n = self.m_mass[-1]
        ret = []
        for i in range(self.m_nt - 2):
            b = self.m0 - sm
            a = m_n + self.m_mass[-i - 2]
            random = tf.random.uniform([n_iter], dtype="float64")
            ms = (b - a) * random + a
            m_n = ms
            sm = sm - self.m_mass[-i - 3]
            ret.append(ms)
        return ret

    def generate(self, n_iter: int, force=True, flatten=True) -> list:
        """generate `n_iter` events

        :param n_iter: number of events
        :param force: switch for cutting generated data to required size
        :param flatten: switch for sampling with weights

        :return:   daughters 4-momentum, list of ndarray with shape (n_iter, 4)
        """
        n_gen = 0
        n_total = n_iter

        mass = self.generate_mass(n_iter)
        if not flatten or self.m_nt == 2:
            pi = self.generate_momentum(mass, n_iter)
            if flatten:
                return pi
            weight = self.get_weight(mass)
            return weight, pi

        mass_f = self.flatten_mass(mass)
        n_gen += int(mass_f[0].shape[0])

        # loop until number of generated events above required
        while force and n_gen < n_iter:
            # guess the total events required
            n_iter2 = int(1.01 * (n_total - n_gen) / (n_gen + 1) * n_iter)
            n_iter2 = min(n_iter2, 4000000)
            mass2 = self.generate_mass(n_iter2)
            mass_f2 = self.flatten_mass(mass2)
            n_gen += mass_f2[0].shape[0]
            n_total += n_iter2
            mass_f = [tf.concat([i, j], 0) for i, j in zip(mass_f, mass_f2)]

        if force:
            mass_f = [i[:n_iter] for i in mass_f]
        return self.generate_momentum(mass_f)

    def generate_momentum(self, mass, n_iter=None):
        """generate random momentum from mass, boost them to a same rest frame"""
        if n_iter is None:
            n_iter = mass[0].shape[0]
        mass_t = [self.m_mass[-1]]
        for i in mass:
            mass_t.append(i)
        mass_t.append(self.m0)
        zeros = tf.zeros([n_iter], dtype="float64")
        p_list = [
            tf.stack([zeros + self.m_mass[-1], zeros, zeros, zeros], axis=-1)
        ]
        for i in range(0, self.m_nt - 1):
            p_list = self.generate_momentum_i(
                mass_t[i + 1], mass_t[i], self.m_mass[-i - 2], n_iter, p_list
            )

        return p_list

    def generate_momentum_i(self, m0, m1, m2, n_iter, p_list=[]):
        """
        :math:`|p|` =  m0,m1,m2 in m0 rest frame
        :param p_list: extra list for momentum need to boost
        """
        # random angle
        cos_theta = 2 * tf.random.uniform([n_iter], dtype="float64") - 1
        sin_theta = tf.sqrt(1 - cos_theta * cos_theta)
        phi = 2 * pi * tf.random.uniform([n_iter], dtype="float64")
        # 4-momentum
        q = tf.broadcast_to(get_p(m0, m1, m2), phi.shape)
        p_0 = tf.sqrt(q * q + m2 * m2)
        p_x = q * sin_theta * tf.cos(phi)
        p_y = q * sin_theta * tf.sin(phi)
        p_z = q * cos_theta

        p = tf.stack([p_0, p_x, p_y, p_z], axis=-1)
        ret = [p]
        # recoil momentum
        p_boost = tf.stack([tf.sqrt(q * q + m1 * m1), p_x, p_y, p_z], axis=-1)
        for i in p_list:
            ret.append(LorentzVector.rest_vector(p_boost, i))
        return ret

    def flatten_mass(self, ms):
        """sampling from mass with weight"""
        weight = self.get_weight(ms)
        rnd = tf.random.uniform(weight.shape, dtype="float64")
        select = weight > rnd
        return [tf.boolean_mask(i, select) for i in ms]

    def get_weight(self, ms):
        r"""calculate weight of mass

        .. math::
            w = \frac{1}{w_{max}} \frac{1}{M}\prod_{i=0}^{n-2} q(M_i,M_{i+1},m_{i+1})

        """
        mass_t = [self.m_mass[-1]]
        for i in ms:
            mass_t.append(i)
        mass_t.append(self.m0)
        R = []
        for i in range(self.m_nt - 1):
            p = get_p(mass_t[i + 1], mass_t[i], self.m_mass[-i - 2])
            R.append(p)
        wt = tf.math.reduce_prod(tf.stack(R), 0)
        return wt / self.m_wtMax

    def set_decay(self, m0, mass):
        r"""set decay mass, calculate max weight

        .. math::
            w_{max} = \frac{1}{M}\prod_{i=0}^{n-2} q(max(M_i),min(M_{i+1}),m_{i+1})

        .. math::
            max(M_i) = M_0 - \sum_{j=1}^{i} (m_j)

        .. math::
            min(M_i) = \sum_{j=i}^{n} (m_j)

        """
        self.m0 = m0
        self.m_nt = len(mass)
        self.m_teCmTm = m0
        for i in mass:
            self.m_mass.append(i)
            self.m_teCmTm = self.m_teCmTm - i

        if self.m_teCmTm <= 0:
            raise ValueError("mass is not validated.")
        emmax = self.m_teCmTm + self.m_mass[-1]
        emmin = 0
        wtmax = 1
        for n in range(1, self.m_nt):
            emmin += self.m_mass[-n]
            emmax += self.m_mass[-n - 1]
            p = get_p(emmax, emmin, self.m_mass[-n - 1])
            wtmax *= p
        self.m_wtMax = tf.convert_to_tensor(wtmax, dtype="float64")


def _get_generator(struct):
    ret = []
    idxs = []

    # depth first flatten
    def _flatten(s, idx=()):
        m0, mi = s
        mi_real = []
        for i, m_i in enumerate(mi):
            if isinstance(m_i, (tuple, list)):
                tmp = m_i[0]
                _flatten(m_i, idx=(*idx, i))
            else:
                tmp = m_i
            mi_real.append(tmp)
        ret.append((m0, mi_real))
        idxs.append(idx)

    _flatten(struct)

    gen = [PhaseSpaceGenerator(m0, mi) for m0, mi in ret]
    return idxs, gen


def _restruct_pi(struct, idxs, pi_s):
    def empty_pi(struct):
        m0, mi = struct
        mi_real = []
        for i, m_i in enumerate(mi):
            if isinstance(m_i, (tuple, list)):
                tmp = empty_pi(m_i)
            else:
                tmp = [None]
            mi_real.append(tmp)
        return [[None], mi_real]

    ret = empty_pi(struct)

    def loop_index(tree, idx):
        for i in idx:
            tree = tree[1][i]
        return tree

    iters = list(zip(idxs, pi_s))
    for idx, pi in iters[::-1]:
        head, tree = loop_index(ret, idx)
        for i, j in zip(tree, pi):
            if isinstance(i[0], list):
                i[0][0] = j
            else:
                i[0] = j

    def tree_boost(p0, tree):
        if isinstance(tree, list):
            return [tree_boost(p0, i) for i in tree]
        if p0 is not None:
            return LorentzVector.rest_vector(LorentzVector.neg(p0), tree)
        return tree

    for idx in idxs:
        all_tree = loop_index(ret, idx)
        head, tree = all_tree
        all_tree[1] = tree_boost(head[0], tree)
        # print(all_tree[0], sum(i[0] if not isinstance(i[0],list) else i[0][0] for i in all_tree[1]))

    def strip_tree(tree):
        if len(tree) == 1:
            return tree[0]
        m1, mi = tree
        return [strip_tree(i) for i in mi]

    return strip_tree(ret)


def generate_phsp(m0, mi, N=1000):
    """general method to generate decay chain phase sapce
    >>> (a, b), c = generate_phsp(1.0, (
    ...                                 (0.3, (0.1, 0.1)),
    ...                                  0.2),
    ...                            N = 10)
    >>> assert np.allclose(LorentzVector.M(a+b+c), 1.0)

    """
    struct = (m0, mi)
    idxs, gen = _get_generator(struct)
    pi = [i.generate(N) for i in gen]
    return _restruct_pi(struct, idxs, pi)



"""
This module implements three classes **Vector3**, **LorentzVector**, **EulerAngle** .
"""
from .tensorflow_wrapper import numpy_cross, tf

_epsilon = 1.0e-14


# import functools
# from pysnooper import snoop


class Vector3(tf.Tensor):
    """
    This class provides methods for 3-d vectors (X,Y,Z)
    """

    def get_X(self):
        return self[..., 0]

    def get_Y(self):
        return self[..., 1]

    def get_Z(self):
        return self[..., 2]

    def norm2(self):
        """
        The norm square
        """
        return tf.reduce_sum(self * self, axis=-1)

    def norm(self):
        return tf.norm(self, axis=-1)

    def dot(self, other):
        """
        Dot product with another Vector3 object
        """
        ret = tf.reduce_sum(self * other, axis=-1)
        return ret

    def cross(self, other):
        """
        Cross product with another Vector3 instance
        """
        p = numpy_cross(self, other)
        return p

    def unit(self):
        """
        The unit vector of itself. It has interface to *tf.linalg.normalize()*.
        """
        p, _n = tf.linalg.normalize(self, axis=-1)
        return p

    def cross_unit(self, other):
        """
        The unit vector of the cross product with another Vector3 object. It has interface to *tf.linalg.normalize()*.
        """
        cro = numpy_cross(self, other)
        norm_cro = tf.expand_dims(tf.norm(cro, axis=-1), -1)
        mask = norm_cro < _epsilon
        bias_other = tf.ones_like(norm_cro) + other
        cro = tf.where(mask, numpy_cross(self, bias_other), cro)
        p, _n = tf.linalg.normalize(cro, axis=-1)
        return p

    def angle_from(self, x, y):
        """
        The angle from x-axis providing the x,y axis to define a 3-d coordinate.

        :param x: A Vector3 instance as x-axis
        :param y: A Vector3 instance as y-axis. It should be perpendicular to the x-axis.
        """
        return tf.math.atan2(Vector3.dot(self, y), Vector3.dot(self, x))

    def cos_theta(self, other):
        """
        cos theta of included angle
        """
        d = Vector3.dot(self, other)
        return d / Vector3.norm(self) / Vector3.norm(other)


class LorentzVector(tf.Tensor):
    """
    This class provides methods for Lorentz vectors (T,X,Y,Z). or -T???
    """

    @staticmethod
    def from_p4(p_0, p_1, p_2, p_3):
        """
        Given **p_0** is a real number, it will make it transform into the same shape with **p_1**.
        """
        zeros = tf.zeros_like(p_1)
        return tf.stack([p_0 + zeros, p_1, p_2, p_3], axis=-1)

    def get_X(self):
        return self[..., 1]

    def get_Y(self):
        return self[..., 2]

    def get_Z(self):
        return self[..., 3]

    def get_T(self):
        return self[..., 0]

    def get_e(self):
        """rm???"""
        return self[..., 0]

    def boost_vector(self):
        """
        :math:`\\beta=(X,Y,Z)/T`
        :return: 3-d vector :math:`\\beta`
        """
        return self[..., 1:4] / self[..., 0:1]

    def vect(self):
        """
        It returns the 3-d vector (X,Y,Z).
        """
        return self[..., 1:4]

    def rest_vector(self, other):
        """
        Boost another Lorentz vector into the rest frame of :math:`\\beta`.
        """
        p = -LorentzVector.boost_vector(self)
        ret = LorentzVector.boost(other, p)
        return ret

    def boost(self, p):
        """
        Boost this Lorentz vector into the frame indicated by the 3-d vector p.
        """
        # pb = Vector3(p)
        pb = p
        beta2 = Vector3.norm2(pb)
        gamma = 1.0 / tf.sqrt(1 - beta2)
        bp = Vector3.dot(pb, LorentzVector.vect(self))
        gamma2 = tf.where(beta2 > _epsilon, (gamma - 1.0) / beta2, 0.0)
        p_r = LorentzVector.vect(self)
        p_r += tf.expand_dims(gamma2 * bp, axis=-1) * pb
        p_r += tf.expand_dims(gamma * LorentzVector.get_T(self), axis=-1) * pb
        T_r = tf.expand_dims(gamma * (LorentzVector.get_T(self) + bp), axis=-1)
        ret = tf.concat([T_r, p_r], -1)
        return ret

    def gamma(self):
        pb = LorentzVector.boost_vector(self)
        beta2 = Vector3.norm2(pb)
        beta2 = tf.where(beta2 < 1, beta2, tf.zeros_like(beta2))
        gamma = 1.0 / tf.sqrt(1 - beta2)
        return gamma

    def beta(self):
        pb = LorentzVector.boost_vector(self)
        beta = Vector3.norm(pb)
        return beta

    def omega(self):
        gamma = LorentzVector.gamma(self)
        omega = tf.acosh(gamma)
        return omega

    def get_metric(self):
        """
        The metric is (1,-1,-1,-1) by default
        """
        return tf.cast(tf.constant([1.0, -1.0, -1.0, -1.0]), self.dtype)

    def M2(self):
        """
        The invariant mass squared
        """
        s = self * self * LorentzVector.get_metric(self)
        return tf.reduce_sum(s, axis=-1)

    def M(self):
        """
        The invariant mass
        """
        m2 = LorentzVector.M2(self)
        return tf.sqrt(tf.abs(m2))

    def neg(self):
        """
        The negative vector
        """
        return tf.concat([self[..., 0:1], -self[..., 1:4]], axis=-1)


class EulerAngle(dict):
    """
    This class provides methods for Eular angle :math:`(\\alpha,\\beta,\\gamma)`
    """

    def __init__(self, alpha=0.0, beta=0.0, gamma=0.0):
        super(EulerAngle, self).__init__()
        self["alpha"] = alpha
        self["beta"] = beta
        self["gamma"] = gamma

    @staticmethod
    def angle_zx_zx(z1, x1, z2, x2):
        """
        The Euler angle from coordinate 1 to coordinate 2 (right-hand coordinates).

        :param z1: Vector3 z-axis of the initial coordinate
        :param x1: Vector3 x-axis of the initial coordinate
        :param z2: Vector3 z-axis of the final coordinate
        :param x2: Vector3 x-axis of the final coordinate
        :return: Euler Angle object
        """
        u_z1 = Vector3.unit(z1)
        u_z2 = Vector3.unit(z2)
        u_y1 = Vector3.cross_unit(z1, x1)
        u_x1 = Vector3.cross_unit(u_y1, z1)
        u_yr = Vector3.cross_unit(z1, z2)
        u_xr = Vector3.cross_unit(u_yr, z1)
        u_y2 = Vector3.cross_unit(z2, x2)
        u_x2 = Vector3.cross_unit(u_y2, z2)
        # np.arctan2(u_xr.Dot(u_y1),u_xr.Dot(u_x1))
        alpha = Vector3.angle_from(u_xr, u_x1, u_y1)
        # np.arctan2(u_z2.Dot(u_xr),u_z2.Dot(u_z1))
        beta = Vector3.angle_from(u_z2, u_z1, u_xr)
        # np.arctan2(u_xr.Dot(u_y2),u_xr.Dot(u_x2))
        gamma = -Vector3.angle_from(u_yr, u_y2, -u_x2)
        return EulerAngle(alpha, beta, gamma)

    @staticmethod
    # @pysnooper.snoop()
    def angle_zx_z_getx(z1, x1, z2):
        """
        The Eular angle from coordinate 1 to coordinate 2. Only the z-axis is provided for coordinate 2, so
        :math:`\\gamma` is set to be 0.

        :param z1: Vector3 z-axis of the initial coordinate
        :param x1: Vector3 x-axis of the initial coordinate
        :param z2: Vector3 z-axis of the final coordinate
        :return eular_angle: EularAngle object with :math:`\\gamma=0`.
        :return x2: Vector3 object, which is the x-axis of the final coordinate when :math:`\\gamma=0`.
        """
        u_z1 = Vector3.unit(z1)
        u_z2 = Vector3.unit(z2)
        u_y1 = Vector3.cross_unit(z1, x1)
        u_x1 = Vector3.cross_unit(u_y1, z1)
        u_yr = Vector3.cross_unit(z1, z2)
        u_xr = Vector3.cross_unit(u_yr, z1)
        # np.arctan2(u_xr.Dot(u_y1),u_xr.Dot(u_x1))
        alpha = Vector3.angle_from(u_xr, u_x1, u_y1)
        # np.arctan2(u_z2.Dot(u_xr),u_z2.Dot(u_z1))
        beta = Vector3.angle_from(u_z2, u_z1, u_xr)
        gamma = tf.zeros_like(beta)
        u_x2 = Vector3.cross_unit(u_yr, u_z2)
        return (EulerAngle(alpha, beta, gamma), u_x2)

    @staticmethod
    def angle_zx_zzz_getx(z, x, zi):
        """
        The Eular angle from coordinate 1 to coordinate 2.
        Z-axis of coordinate 2 is the normal vector of a plane.

        :param z1: Vector3 z-axis of the initial coordinate
        :param x1: Vector3 x-axis of the initial coordinate
        :param z: list of Vector3 of the plane point.
        :return eular_angle: EularAngle object.
        :return x2: list of Vector3 object, which is the x-axis of the final coordinate in zi.
        """
        z1, z2, z3 = zi
        zz = Vector3.cross_unit(z1 - z2, z2 - z3)
        xi = [Vector3.cross_unit(i, zz) for i in zi]
        ang = EulerAngle.angle_zx_zx(z, x, zz, xi[2])
        return ang, xi


class SU2M(dict):
    def __init__(self, x):
        self["x"] = x

    @staticmethod
    def Boost_z(omega):
        zeros = tf.zeros_like(omega)
        a = tf.complex(tf.exp(omega / 2), zeros)
        zeros_c = tf.complex(zeros, zeros)
        ret = SU2M([[1 / a, zeros_c], [zeros_c, a]])
        return ret

    @staticmethod
    def Boost_z_from_p(p):
        omega = LorentzVector.omega(p)
        return SU2M.Boost_z(omega)

    @staticmethod
    def Rotation_z(alpha):
        zeros = tf.zeros_like(alpha)
        zeros_c = tf.complex(zeros, zeros)
        a = tf.exp(tf.complex(zeros, alpha / 2))
        return SU2M([[1 / a, zeros_c], [zeros_c, a]])

    @staticmethod
    def Rotation_y(beta):
        zeros = tf.zeros_like(beta)
        s = tf.complex(tf.sin(beta / 2), zeros)
        c = tf.complex(tf.cos(beta / 2), zeros)
        return SU2M([[c, -s], [s, c]])

    def __mul__(self, other):
        x = self["x"]
        y = other["x"]
        aa = x[0][0] * y[0][0] + x[0][1] * y[1][0]
        ab = x[0][0] * y[0][1] + x[0][1] * y[1][1]
        ba = x[1][0] * y[0][0] + x[1][1] * y[1][0]
        bb = x[1][0] * y[0][1] + x[1][1] * y[1][1]
        return SU2M([[aa, ab], [ba, bb]])

    def inv(self):
        aa, ab = self["x"][0]
        ba, bb = self["x"][1]
        return SU2M([[bb, -ab], [-ba, aa]])

    def get_euler_angle(self):
        x = self["x"]
        cosbeta = tf.math.real(x[0][0] * x[1][1] + x[0][1] * x[1][0])
        cosbeta = tf.clip_by_value(cosbeta, -1, 1)
        zeros = tf.zeros_like(cosbeta)
        beta = tf.math.acos(cosbeta)
        m_1 = tf.abs(x[0][0])
        m_2 = tf.abs(x[1][0])
        # alpha_p_gamma = tf.math.imag(
        #     tf.math.log(x[1][1] / tf.complex(m_1, zeros))
        # )
        alpha_p_gamma = tf.math.angle(x[1][1])
        alpha_m_gamma = -tf.math.angle(x[1][0])
        # -tf.math.imag(
        #    tf.math.log(x[1][0] / tf.complex(m_2, zeros))
        # )
        alpha = alpha_p_gamma + alpha_m_gamma
        gamma = alpha_p_gamma - alpha_m_gamma
        return EulerAngle(alpha, beta, gamma)

    def __repr__(self):
        return str(tf.stack(self["x"]))


class AlignmentAngle(EulerAngle):
    @staticmethod
    def angle_px_px(p1, x1, p2, x2):
        z1 = LorentzVector.vect(p1)
        zr = LorentzVector.vect(p2 - p1)
        z2 = LorentzVector.vect(p2)
        r1, xr = EulerAngle.angle_zx_z_getx(z1, x1, zr)
        m_p = LorentzVector.M(p1)
        p3_1 = Vector3.norm(LorentzVector.vect(p1))
        p3_2 = Vector3.norm(LorentzVector.vect(p2))
        zeros = tf.zeros_like(p3_1)
        delta_p = p3_2 - p3_1
        p4_2 = LorentzVector.from_p4(
            tf.sqrt(m_p * m_p + delta_p * delta_p), zeros, zeros, delta_p
        )
        b1 = SU2M.Boost_z_from_p(p4_2)
        r2 = EulerAngle.angle_zx_zx(zr, xr, z2, x2)
        r_1 = b1 * SU2M.Rotation_y(r1["beta"]) * SU2M.Rotation_z(r1["alpha"])
        r_2 = (
            SU2M.Rotation_z(r2["gamma"])
            * SU2M.Rotation_y(r2["beta"])
            * SU2M.Rotation_z(r2["alpha"])
        )
        r = r_2 * r_1
        return r.get_eular_angle()


def kine_min_max(s12, m0, m1, m2, m3):
    """ min max s23 for s12 in p0 -> p1 p2 p3"""
    m12 = tf.sqrt(s12)
    m12 = tf.where(m12 > (m1 + m2), m12, m1 + m2)
    m12 = tf.where(m12 < (m0 - m3), m12, m0 - m3)
    # if(mz < (m_d+m_pi)) return 0;
    # if(mz > (m_b-m_pi)) return 0;
    E2st = 0.5 * (m12 * m12 - m1 * m1 + m2 * m2) / m12
    E3st = 0.5 * (m0 * m0 - m12 * m12 - m3 * m3) / m12
    p2st = tf.sqrt(E2st * E2st - m2 * m2)
    p3st = tf.sqrt(E3st * E3st - m3 * m3)
    s_min = (E2st + E3st) ** 2 - (p2st + p3st) ** 2
    s_max = (E2st + E3st) ** 2 - (p2st - p3st) ** 2
    return s_min, s_max


def kine_min(s12, m0, m1, m2, m3):
    """ min s23 for s12 in p0 -> p1 p2 p3"""
    s_min, s_max = kine_min_max(s12, m0, m1, m2, m3)
    return s_min


def kine_max(s12, m0, m1, m2, m3):
    """ max s23 for s12 in p0 -> p1 p2 p3"""
    s_min, s_max = kine_min_max(s12, m0, m1, m2, m3)
    return s_max



import argparse
import os

from jinja2 import Template


def load_config(input_file, variable_file=""):
    """
    read input_file and render it with variable file
    """
    with open(input_file) as f:
        template = Template(f.read())

    variable = os.environ.copy()
    if variable_file != "":
        with open(variable_file) as f:
            exec(f.read(), variable)

    ret = template.render(**variable)
    return ret


def main():
    parser = argparse.ArgumentParser(description="Jinja")
    parser.add_argument("input_file")
    parser.add_argument("variable", nargs="?", default="")
    args = parser.parse_args()

    s = load_config(args.input_file, args.variable)
    print(s)


if __name__ == "__main__":
    main()



"""
scripts for calcaulate split angle amplitude
"""

import json

import matplotlib.pyplot as plt
import numpy as np

from tf_pwa.config_loader import ConfigLoader
from tf_pwa.data import data_index, data_to_numpy
from tf_pwa.experimental.build_amp import build_angle_amp_matrix


def main():
    config = ConfigLoader("config.yml")
    decay = config.get_amplitude().decay_group
    data = config.get_data("phsp")
    ret = []
    for i in data:
        cached_amp = build_angle_amp_matrix(decay, i)
        ret.append(data_to_numpy(cached_amp))

    idx = config.get_data_index("angle", "R_BC/B")
    ang = data_index(data[0], idx)

    np.savez("phsp.npz", ret)

    for k, v in ret[0].items():
        for i, amp in enumerate(v):
            w = np.abs(amp) ** 2
            w = np.sum(np.reshape(w, (amp.shape[0], -1)), axis=-1)
            plt.hist(
                np.cos(ang["beta"]),
                weights=w,
                bins=20,
                histtype="step",
                label="{}: {}".format(k, i),
            )
    plt.savefig("angle_costheta.png")


if __name__ == "__main__":
    main()



import numpy as np

from .utils import error_print


class NumberError(object):
    """ basic class for propagation of error"""

    def __init__(self, value, error=1.0):
        self._value = value
        self._error = error

    def __repr__(self):
        return error_print(self._value, self._error)

    @property
    def value(self):
        return self._value

    @property
    def error(self):
        return self._error

    def __add__(self, other):
        if isinstance(other, NumberError):
            val = self._value + other._value
            err = np.sqrt(self._error ** 2 + other._error ** 2)
        else:
            val = self._value + other
            err = self._error
        return NumberError(val, err)

    def __sub__(self, other):
        if isinstance(other, NumberError):
            val = self._value - other._value
            err = np.sqrt(self._error ** 2 + other._error ** 2)
        else:
            val = self._value - other
            err = self._error
        return NumberError(val, err)

    def __neg__(self):
        val = -self._value
        err = self._error
        return NumberError(val, err)

    def __mul__(self, other):
        if isinstance(other, NumberError):
            val = self._value * other._value
            err = np.sqrt(
                (self._error * other._value) ** 2
                + (self._value * other._error) ** 2
            )
        else:
            val = self._value * other
            err = self._error * other
        return NumberError(val, err)

    def __truediv__(self, other):
        if isinstance(other, NumberError):
            val = self._value / other._value
            err = (
                np.sqrt(
                    (self._error) ** 2
                    + (self._value * other._error / other._value) ** 2
                )
                / other._value
            )
        else:
            val = self._value / other
            err = self._error / other
        return NumberError(val, err)

    def __pow__(self, other):
        if isinstance(other, NumberError):
            val = self._value ** other._value
            err1 = (
                other._value * self._value ** (other._value - 1) * self._error
            )
            err2 = np.log(other._value) * val * other._error
            err = np.sqrt(err1 ** 2 + err2 ** 2)
        else:
            val = self._value ** other
            err = np.abs(other * self._value ** (other - 1)) * self._error
        return NumberError(val, err)

    def __rpow__(self, other):
        val = other ** self._value
        err = np.log(self._value) * val * self._error
        return NumberError(val, err)

    def log(self):
        val = np.log(self._value)
        err = self._error / np.abs(self._value)
        return NumberError(val, err)

    def exp(self):
        val = np.exp(self._value)
        err = val * self._error
        return NumberError(val, err)

    def apply(self, fun, grad=None, dx=1e-5):
        val = fun(self._value)
        if grad is not None:
            err = np.abs(grad(self._value)) * self._error
        else:
            grad_v = (fun(self._value + dx) - fun(self._value - dx)) / 2 / dx
            err = np.abs(grad_v) * self._error
        return NumberError(val, err)


def cal_err(fun, *args, grad=None, dx=1e-5, kwargs=None):
    kwargs = {} if kwargs is None else kwargs
    value = []
    errs = []
    for i in args:
        if isinstance(i, NumberError):
            value.append(i._value)
            errs.append(i._error)
        else:
            value.append(i)
            errs.append(0.0)
    val = fun(*value, **kwargs)
    if grad is None:
        grad_v = []
        for i, k in enumerate(value):
            value[i] += dx
            val_p = fun(*value, **kwargs)
            value[i] -= 2 * dx
            val_m = fun(*value, **kwargs)
            value[i] = k
            grad_v.append((val_p - val_m) / 2 / dx)
    else:
        grad_v = grad(*value, **kwargs)
    err = np.sqrt(sum((i * j) ** 2 for i, j in zip(grad_v, errs)))
    return NumberError(val, err)
